# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

from dingDong.misc.logger     import p
from dingDong.misc.enumsJson import eConn, eJson
from dingDong.config import config

from dingDong.conn.connGlobalDB   import baseGlobalDb
from dingDong.conn.connAccess     import access
from dingDong.conn.connFile       import connFile

CLASS_TO_LOAD = {eConn.SQLSERVER :baseGlobalDb,
                 eConn.ORACLE:baseGlobalDb,
                 eConn.VERTICA:baseGlobalDb,
                 eConn.ACCESS:access,
                 eConn.MYSQL:baseGlobalDb,
                 eConn.LITE:baseGlobalDb,
                 eConn.FILE:connFile}

def addPropToDict (existsDict, newProp):
    if newProp and isinstance(newProp, (dict, OrderedDict)):
        for k in newProp:
            if k in existsDict and isinstance(newProp[k], dict):
                existsDict = addPropToDict (existsDict, newProp=newProp[k])
            elif k not in existsDict:
                existsDict[k] = newProp[k]
            elif k in existsDict and existsDict[k] is None:
                existsDict[k] = newProp[k]
    elif isinstance(newProp,str):
        existsDict[eJson.jValues.URL] = newProp
    else:
        p("THERE IS AN ERROR ADDING %s INTO DICTIONARY " %(newProp), "e")

    return existsDict

def mngConnectors(connPropDic, connLoadProp=None):

    connLoadProp = connLoadProp if connLoadProp else config.CONN_URL

    ## Merge by CONNECTION
    if eJson.jValues.CONN in connPropDic and connPropDic[eJson.jValues.CONN] in connLoadProp:
        connPropDic = addPropToDict(existsDict=connPropDic, newProp=connLoadProp[connPropDic[eJson.jValues.CONN]])
        # update Connection type
        if eJson.jValues.CONN in connLoadProp[connPropDic[eJson.jValues.CONN]] and connLoadProp[connPropDic[eJson.jValues.CONN]][eJson.jValues.CONN] is not None:
            connPropDic[eJson.jValues.CONN] = connLoadProp[connPropDic[eJson.jValues.CONN]][eJson.jValues.CONN]

    if eJson.jValues.NAME in connPropDic and connPropDic[eJson.jValues.NAME] in connLoadProp :
        connPropDic = addPropToDict(existsDict=connPropDic, newProp=connLoadProp[connPropDic[eJson.jValues.NAME]])

    elif eJson.jValues.TYPE in connPropDic and connPropDic[eJson.jValues.TYPE] in connLoadProp :
        connPropDic = addPropToDict(existsDict=connPropDic, newProp=connLoadProp[connPropDic[eJson.jValues.TYPE]])

    if eJson.jValues.NAME not in connPropDic or connPropDic[eJson.jValues.NAME] is None:
        connPropDic[eJson.jValues.NAME] = connPropDic[eJson.jValues.CONN]

    # Tal: if ther is no URL, check if conn is Set, search if there is only one conn and update details
    if  eJson.jValues.URL not in connPropDic or (connPropDic[eJson.jValues.URL] is None and connPropDic[eJson.jValues.CONN] is not None):
        connToLook  = connPropDic[eJson.jValues.CONN]
        connParams  = None
        connNum     = 0

        for c in connLoadProp:
            if c == connToLook or (eJson.jValues.CONN in connLoadProp[c] and c == connLoadProp[c][eJson.jValues.CONN] ):
                connParams = connLoadProp[c]
                connNum+=1
        if connNum == 1:
            for prop in connParams:
                if prop not in connPropDic:
                    connPropDic[prop] = connParams[prop]
                elif connParams[prop] is None:
                    connParams[prop] = connParams[prop]
        else:
            p("FOUND %s used %s times. cannot use it as URL defintion" %(str(connToLook),str(connNum)), "w")
    if connPropDic and isinstance(connPropDic, dict) and eJson.jValues.CONN in connPropDic:
        cType = connPropDic[eJson.jValues.CONN]
        if cType in CLASS_TO_LOAD:
            return CLASS_TO_LOAD[cType]( connPropDict=connPropDic )

        else:
            p("CONNECTION %s is NOT DEFINED. PROP: %s" % (str(cType), str(connPropDic)), "e")
    else:
        p ("connectorMng->mngConnectors: must have TYPE prop. prop: %s " %(str(connPropDic)), "e")

def __queryParsetIntoList(self, sqlScript, getPython=True, removeContent=True, dicProp=None, pythonWord="popEtl"):
        if isinstance(sqlScript, (tuple, list)):
            sqlScript = "".join(sqlScript)
        # return list of sql (splitted by list of params)
        allQueries = self.__getAllQuery(longStr=sqlScript, splitParam=['GO', u';'])

        if getPython:
            allQueries = self.__getPythonParam(allQueries, mWorld=pythonWord)

        if removeContent:
            allQueries = self.__removeComments(allQueries)

        if dicProp:
            allQueries = self._replaceProp(allQueries, dicProp)

        return allQueries


# Copyright (c) 2017-2019, BPMK LTD (BiSkilled) Tal Shany <tal.shany@biSkilled.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import re
import copy
from collections import OrderedDict

from dingDong.config            import config
from dingDong.misc.logger       import p, LOGGER_OBJECT
from dingDong.misc.misc         import uniocdeStr
from dingDong.bl.jsonParser     import jsonParser
from dingDong.misc.enumsJson    import eJson
from dingDong.conn.baseConnectorManager   import mngConnectors as conn

## Execters
from dingDong.executers.executeSql import execQuery
from dingDong.executers.executeAddMsg import executeAddMsg
from dingDong.executers.executeMicrosoftOLAP import OLAP_Process

class dingDong:
    def __init__ (self,  dicObj=None, filePath=None,
                dirData=None, includeFiles=None, notIncludeFiles=None,
                dirLogs=None,connDict=None, processes=None):

        self.jsonParser = jsonParser(dicObj=dicObj, filePath=filePath,
                                     dirData=dirData, includeFiles=includeFiles, notIncludeFiles=notIncludeFiles,
                                     connDict=connDict)

        self.msg = executeAddMsg()
        self.propcesses = processes if processes else config.NUM_OF_PROCESSES

        if dirLogs or config.LOGS_DIR:
            LOGGER_OBJECT.setLogsFiles (logDir=dirLogs)


        ## Defualt properties
        self.connDict       = self.jsonParser.connDict
        self.setDefaults()

    def setDefaults (self):
        self.stt = None
        self.addSourceColumn = True
        self.src = None
        self.tar = None
        self.onlyTarget = False
        self.mrg = None
        self.execProc = None

        self.mergeSource = None
        self.mergeTarget = None
        self.mergeKeys = None

    def ding (self, destList=None, jsName=None, jsonNodes=None):
        p('STARTING TO MODEL DATA STRUCURE >>>>>' , "i")
        if jsName and jsonNodes:
            allNodes = [(jsName, jsonNodes)]
        else:
            allNodes = [(jsName, jsonNodes) for jsName, jsonNodes in self.__getNodes(destList=destList)]

        for jsName, jsonNodes in allNodes:
            self.setDefaults()
            for jMap in jsonNodes:
                self.__setSTTandOnlyTarget(node=jMap)
                for node in jMap:
                    if isinstance(jMap[node], (list, tuple)):
                        self.ding(destList=None, jsName=jsName, jsonNodes=jMap[node])
                        continue

                    if not isinstance(jMap[node], (dict, OrderedDict)):
                        p("Not valid json values - must be list of dictionary.. continue. val: %s " % (str(jMap)),"e")
                        continue

                    self.__setMainProperty(key=node, valDict=jMap[node])

                    if self.src:
                        mergeSource = copy.copy(self.src)

                    if self.tar:
                        mergeSource = copy.copy(self.tar)

                        if self.onlyTarget:
                            self.tar.create(stt=self.stt)
                            self.tar.close()
                            self.tar = None

                    if self.tar and self.src:
                        # convert source data type to target data types
                        targetStt = self.__updateTargetBySourceAndStt(src=self.src, tar=self.tar)

                        self.tar.create(stt=targetStt)
                        mergeSource = copy.copy(self.tar)
                        self.src.close()
                        self.tar.close()

                        self.src = None
                        self.tar = None


                    if self.mrg and mergeSource:
                        sttMerge = self.__updateTargetBySourceAndStt(src=mergeSource, tar=self.mrg)
                        self.mrg.create(stt=sttMerge)
                        self.mrg.close()
                        self.mrg = None


        p('FINSHED TO MODEL DATA STRUCURE >>>>>', "i")
        p('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', "ii")

    def dong (self, destList=None, jsName=None, jsonNodes=None):
        p('STARTING TO TRANSFER DATA STRUCURE >>>>>', "i")
        if jsName and jsonNodes:
            allNodes = [(jsName, jsonNodes)]
        else:
            allNodes = [(jsName, jsonNodes) for jsName, jsonNodes in self.__getNodes(destList=destList)]

        for jsName, jsonNodes in allNodes:
            self.src = None
            self.tar = None
            self.mrg = None
            self.execProc = None
            self.stt = None
            self.addSourceColumn = True

            for jMap in jsonNodes:
                self.__setSTTandOnlyTarget(node=jMap)

                for node in jMap:
                    if isinstance(jMap[node], (list, tuple)):
                        self.dong(destList=destList, jsName=jsName, jsonNodes=jMap[node])
                        continue

                    if not isinstance(jMap[node], (dict, OrderedDict)):
                        p("Not valid json values - must be list of dictionary.. continue. val: %s " % (str(jMap)),"e")
                        continue

                    self.__setMainProperty(key=node, valDict=jMap[node])

                    if self.execProc:
                        """ Execute internal connection procedure """
                        self.execProc = None

                    if self.src and self.tar:
                        """ TRANSFER DATA FROM SOURCE TO TARGET """
                        self.tar.preLoading()

                        tarToSrc = self.__mappingLoadingSourceToTarget(src=self.src, tar=self.tar)
                        self.src.extract(tar=self.tar, tarToSrc=tarToSrc, addAsTaret=True)
                        self.tar.close()
                        self.src.close()

                        self.src = None
                        self.tar = None

                    """ MERGE DATA BETWEEN SOURCE AND TARGET TABLES """
                    if self.mrg and self.mergeSource:
                        self.mrg.merge(mergeTable=self.mergeTarget, mergeKeys=self.mergeKeys, sourceTable=None)
                        self.mrg.close()

                        self.mergeSource = None
                        self.mergeTarget = None
                        self.mergeKeys = None

        p('FINISHED TO TRANSFER DATA STRUCURE >>>>>', "i")
        p('>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>', "ii")

    def setLoggingLevel (self, val):
        CRITICAL = 50
        ERROR = 40
        WARNING = 30
        INFO = 20
        DEBUG = 10
        NOTSET = 0

        if val in (CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET):
            config.LOGS_DEBUG = val
            LOGGER_OBJECT.logLevel = val
            LOGGER_OBJECT.setLogLevel(logLevel=config.LOGS_DEBUG)
        else:
            err = "Logging is not valid, valid values: 0,10,20,30,40,50"
            raise ValueError(err)

    """ LOADING BUSINESS LOGIC EXECUTOERS """
    def execDbSql(self, queries, connName=None, connType=None, connUrl=None, connPropDic=None):
        connPropDic = connPropDic if connPropDic else {}
        if connName : connPropDic[eJson.jValues.NAME] = connName
        if connType : connPropDic[eJson.jValues.CONN] = connType
        if connUrl  : connPropDic[eJson.jValues.URL] = connUrl

        connObj = conn(connPropDic=connPropDic , connLoadProp=self.connDict)
        execQuery(sqlWithParamList=queries, connObj=connObj)

    def execMicrosoftOLAP (self, serverName, dbName, cubes=[], dims=[], fullProcess=True):
        OLAP_Process(serverName=serverName, dbName=dbName, cubes=cubes, dims=dims, fullProcess=fullProcess)

    """ Check Source values in STT - remove invalid values """
    def __updateSTTBySource (self, srcStructure, pre="[", pos="]"):
        # Check if ther are sourcea in STT that not defined
        srcStrucureL = {x.replace(pre,"").replace(pos,"").lower():x for x in srcStructure}

        removeColumns = []
        for col in self.stt:
            if eJson.jSttValues.SOURCE in self.stt[col] and self.stt[col][eJson.jSttValues.SOURCE].replace(pre,"").replace(pos,"").lower() not in srcStrucureL:
                removeColumns.append (col)

        for col in removeColumns:
            p("STT TAREGT %s HAVE INVALID SOURCE %s --> ignore COLUMN " % (col, self.stt[col][eJson.jSttValues.SOURCE]),"w")
            del self.stt[col]

    """ Check Taret values in STT - remove invalid values  """
    def __updateSTTByTarget (self, tarStructure, pre="[", pos="]"):
        # Check if ther are sourcea in STT that not defined
        tarStrucureL = {x.replace(pre,"").replace(pos,"").lower():x for x in tarStructure}
        for col in self.stt:
            if col.replace(pre,"").replace(pos,"").lower() not in tarStrucureL:
                p("STT COLUMN %s NOT EXISTS OBJECT --> ignore COLUMN " %(col) ,"w")
                #del self.stt[col]

    def __setSTTandOnlyTarget (self, node):
        if eJson.jKeys.STTONLY in node:
            self.stt            = node[eJson.jKeys.STTONLY]
            self.addSourceColumn= False

        elif eJson.jKeys.STT in node:
            self.stt = node[eJson.jKeys.STT]
        else:
            self.stt            = {}
            self.addSourceColumn= True

        if eJson.jKeys.SOURCE not in node and eJson.jKeys.QUERY not in node and self.stt is not None:
            self.onlyTarget = True

    def __setMainProperty (self,key, valDict):
        if eJson.jKeys.EXEC == key or eJson.jKeys.EXEC in valDict:
            self.execProc = valDict

        elif eJson.jKeys.SOURCE == key or eJson.jKeys.SOURCE in valDict:
            valDict[eJson.jValues.IS_SOURCE] = True
            self.src = conn(connPropDic=valDict)

        elif eJson.jKeys.QUERY == key or eJson.jKeys.QUERY in valDict:
            valDict[eJson.jValues.IS_SOURCE] = True
            self.src = conn(connPropDic=valDict)

        elif eJson.jKeys.TARGET == key or eJson.jKeys.TARGET in valDict:
            valDict[eJson.jValues.IS_TARGET] = True
            self.tar = conn(connPropDic=valDict)

        elif eJson.jKeys.MERGE == key or eJson.jKeys.MERGE in valDict:
            self.mergeSource= valDict[eJson.jMergeValues.SOURCE]
            self.mergeTarget= valDict[eJson.jMergeValues.TARGET]
            self.mergeKeys  = valDict[eJson.jMergeValues.MERGE]

            if self.tar and self.tar.connObj == self.mergeSource:
                self.mrg = self.tar
            elif self.src and self.src.connObj == self.mergeSource:
                self.mrg = self.src
            else:
                p("MERGE SOURCE %s IS NOT FOUND, IGNORE MERGE " %(self.mergeSource),"e")
        elif eJson.jKeys.STTONLY == key or eJson.jKeys.STT == key:
            pass
        else:
            p("IGNORE NODE: KEY:%s, VALUE:%s " % (key,str(valDict)), "e")


    """ LOADING MODULE: Return mapping between source and target """
    def __mappingLoadingSourceToTarget (self, src, tar):
        tarToSrc        = OrderedDict()
        srcStructure    = src.getStructure()
        tarStructure    = tar.getStructure()

        srcPre, srcPos,tarPre, tarPos = '','','',''
        if hasattr(src, 'columnFrame'):
            srcPre, srcPos = src.columnFrame[0], src.columnFrame[1]

        if hasattr(tar, 'columnFrame'):
            tarPre, tarPos = tar.columnFrame[0], tar.columnFrame[1]


        # remove from STT column that not exists in Target OBJECT OR Source OBJECT
        if tar.usingSchema:
            self.__updateSTTByTarget(tarStructure=tarStructure, pre=tarPre, pos=tarPos)
        self.__updateSTTBySource(srcStructure=srcStructure, pre=srcPre, pos=srcPos)

        srcColumns = OrderedDict()
        tarColumns = OrderedDict({x.replace(tarPre, "").replace(tarPos, "").lower(): x for x in tarStructure})
        sttColumns = OrderedDict({x.replace(tarPre, "").replace(tarPos, "").lower(): x for x in self.stt})


        ## {srcName in Target: Source column }
        for col in srcStructure:
            colAlias= srcStructure[col][eJson.jSttValues.ALIACE] if eJson.jSttValues.ALIACE in srcStructure[col] else None
            colName = colAlias if colAlias else col
            srcColumns[colName.replace(srcPre, "").replace(srcPos, "").lower()] = col

        # There is no target schema --> using all source and STT
        if tar.usingSchema and self.addSourceColumn:
            for col in srcColumns:
                tarToSrc[col] = {eJson.jSttValues.SOURCE: srcColumns[col]}

        # ADD ALL COLUMN FROM SOURCE
        elif self.addSourceColumn:
            for col in tarColumns:
                if col in srcColumns:
                    tarToSrc[tarColumns[col]] = {eJson.jSttValues.SOURCE: srcColumns[col]}

        tarToSrcColumns = {x.replace(tarPre, "").replace(tarPos, "").lower(): x for x in tarToSrc}
        for col in sttColumns:
            if col in tarToSrcColumns:
                tarToSrc[tarToSrcColumns[col]].update(self.stt[sttColumns[col]])
            else:
                tarToSrc[sttColumns[col]] = self.stt[sttColumns[col]]

        if not tar.usingSchema:
            p("TAREGT %s DO NOT HAVE FIX SCHEMA " %(tar.conn), "ii")
            return tarToSrc

        #### Check Column in Source and not exists in mapping
        existsTarColumns = {}
        existsSrcColumns = {}

        for col in tarToSrc:
            existsTarColumns[col.replace(tarPre, "").replace(tarPos, "").lower()] = col
            if eJson.jSttValues.SOURCE in tarToSrc[col] and tarToSrc[col][eJson.jSttValues.SOURCE]:
                existsSrcColumns[
                    tarToSrc[col][eJson.jSttValues.SOURCE].replace(srcPre, "").replace(srcPos, "").lower()] = \
                tarToSrc[col][eJson.jSttValues.SOURCE]

        columnNotMapped = u""
        for col in tarColumns:
            if col not in existsTarColumns:
                columnNotMapped+=uniocdeStr(tarColumns[col])+u" ; "
        if len(columnNotMapped) > 0:
            p(u"TARGET COLUMN NOT MAPPED: %s" % (columnNotMapped), "w")

        columnNotMapped = u""
        for col in srcColumns:
            if srcColumns[col].replace(srcPre, "").replace(srcPos, "").lower() not in existsSrcColumns:
                columnNotMapped+=uniocdeStr(srcColumns[col])+u" ; "
        if len(columnNotMapped) > 0:
            p(u"SOURCE COLUMN NOT MAPPED: %s" % (columnNotMapped), "w")

        return tarToSrc

    """ MAPPING MODULE: Convert Source Data Type into Target Data Type """
    def __updateTargetBySourceAndStt(self, src, tar):
        retStrucure = OrderedDict()
        sourceStt   = src.getStructure()
        sourceColL  = {x.lower():x for x in sourceStt}

        if src.conn == tar.conn:
            if self.addSourceColumn:
                for col in sourceStt:
                    if eJson.jSttValues.ALIACE in sourceStt[col] and sourceStt[col][eJson.jSttValues.ALIACE]:
                        retStrucure[ sourceStt[col][eJson.jSttValues.ALIACE] ] = {eJson.jSttValues.TYPE: sourceStt[col][eJson.jSttValues.TYPE]}
                    else:
                        retStrucure[col] = {eJson.jSttValues.TYPE: sourceStt[col][eJson.jSttValues.TYPE]}

        ### Source connection type is different than target connection type
        elif self.addSourceColumn:

            for col in sourceStt:
                targetColName = col
                if eJson.jSttValues.ALIACE in sourceStt[col] and sourceStt[col][eJson.jSttValues.ALIACE]:
                    targetColName = sourceStt[col][eJson.jSttValues.ALIACE]

                colType = sourceStt[col][eJson.jSttValues.TYPE] if eJson.jSttValues.TYPE in sourceStt[col] and sourceStt[col][eJson.jSttValues.TYPE] else tar.defDataType
                fmatch = re.search(r'(.*)(\(.+\))', colType, re.M | re.I)
                if fmatch:
                    replaceString   = fmatch.group(1)  # --> varchar, int , ...
                    postType        = fmatch.group(2)  # --> (X), (X,X) , ....
                else:
                    replaceString   = colType
                    postType        = ''

                ## Receive list of all dataType in DataTypes Tree
                newDataTypeTree     = src.getDataTypeTree (dataType=replaceString.lower(), ret=([]))
                if newDataTypeTree is None:
                    p("SOURCE CONNECTION: %s, COLUMN: %s, DATA TYPE: %s ; IS NOT EXISTS, WILL USE DEFAULT VALUE" %(src.conn,col, replaceString),"w")
                    tarType = tar.defDataType
                else:
                    targetList = tar.setDataTypeTree (dataTypeTree=newDataTypeTree, allDataTypes=tar.DATA_TYPES, ret=[])
                    tarType = '%s%s' %(targetList[-1],postType) if targetList and len(targetList)>0 else tar.defDatatType
                retStrucure[targetColName] = {eJson.jSttValues.TYPE:tarType}

        retStrucureL = {x.lower():x for x in retStrucure}
        retStrucureKeys = list (retStrucure.keys())

        for col in self.stt:
            if eJson.jSttValues.SOURCE in self.stt[col]:
                sourceName = self.stt[col][eJson.jSttValues.SOURCE]
                if sourceName in retStrucureKeys:
                    del retStrucure[sourceName]
                elif sourceName.lower() in retStrucureL:
                    del retStrucure[retStrucureL[sourceName.lower()]]

            if col.lower() in sourceColL and eJson.jSttValues.TYPE in self.stt[col] and self.stt[col][eJson.jSttValues.TYPE]:
                retStrucure[sourceColL[ col.lower() ]][eJson.jSttValues.TYPE] = self.stt[col][eJson.jSttValues.TYPE]

                if eJson.jSttValues.ALIACE in self.stt[col]:
                    retStrucure[sourceColL[col.lower()]][eJson.jSttValues.ALIACE] = self.stt[col][
                        eJson.jSttValues.ALIACE]

            elif col.lower() not in sourceColL:
                if eJson.jSttValues.TYPE in self.stt[col] and self.stt[col][eJson.jSttValues.TYPE]:
                    retStrucure[col] = {eJson.jSttValues.TYPE: self.stt[col][eJson.jSttValues.TYPE]}
                else:
                    retStrucure[col] = {eJson.jSttValues.TYPE: self.tar.defDataType}

                    if eJson.jSttValues.ALIACE in self.stt[col]:
                        retStrucure[col][eJson.jSttValues.ALIACE] = self.stt[col][eJson.jSttValues.ALIACE]
        return retStrucure

    def __getNodes(self, destList=None):
        for jsonMapping in self.jsonParser.jsonMappings(destList=destList):
            for jsName in jsonMapping:
                yield jsName, jsonMapping[jsName]


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import abc
import six
import sys
import re
import copy


from dingDong.misc.logger   import p
from dingDong.misc.enumsJson import eConn, eJson, findProp
from dingDong.misc.misc     import replaceStr, uniocdeStr


DEFAULTS    =   {
                    eJson.jValues.DEFAULT_TYPE:eConn.dataType.B_STR,
                    eJson.jValues.BATCH_SIZE:200000
                }

DATA_TYPES = {
    eConn.dataType.B_STR: { eConn.dataType.B_DEFAULT:'nvarchar(200)',
                            eConn.dataType.DB_VARCHAR:None,
                            eConn.dataType.DB_NVARCHAR:None,
                            eConn.dataType.DB_CHAR:None,
                            eConn.dataType.DB_BLOB:None},
    eConn.dataType.B_INT: {eConn.dataType.DB_INT:None,
                           eConn.dataType.DB_BIGINT:None},
    eConn.dataType.B_FLOAT:{eConn.dataType.DB_FLOAT:None,
                            eConn.dataType.DB_DECIMAL:None},
    eConn.dataType.DB_DATE:{eConn.dataType.DB_DATE:None}
}

""" baseConn -- get connection propertirs : conn, connUrl,  connExtraUrl, connName,connDataTypes, connDefaultType, connPropDic"""
@six.add_metaclass(abc.ABCMeta)
class baseBatch ():
    def __init__ (self, conn=None, connName=None, connPropDict=None, defaults=None, dataType=None, update=None):
        self.connPropDict   = connPropDict
        self.conn           = self.setProperties (propKey=eJson.jValues.CONN, propVal=conn)
        self.connName       = self.setProperties (propKey=eJson.jValues.NAME, propVal=connName)
        self.usingSchema    = True
        self.update         = self.setProperties (propKey=eJson.jValues.UPDATE, propVal=update, propDef=-1 )

        if not self.conn:
            self.conn = self.connName


        self.baseDefaults   = DEFAULTS
        self.baseDataTypes  = copy.deepcopy(DATA_TYPES)
        self.baseDefDataType = self.baseDefaults[eJson.jValues.DEFAULT_TYPE]

        self.DEFAULTS   = self.setDefaults(defaultsDic=defaults)
        self.defDataType= self.DEFAULTS[eJson.jValues.DEFAULT_TYPE]

        self.DATA_TYPES = self.setDataTypes(connDataTypes=dataType)
        self.batchSize  = self.DEFAULTS[eJson.jValues.BATCH_SIZE]

        if not findProp (prop=self.conn, obj=eConn):
            err  = "Connection type is not valid: %s, use valid connection properties" %(str(self.conn))
            raise ValueError(err)

    """ ABSTRACT METHOD THAT MUST BE IMPLEMETED IN ANY CHILD CLASSES """
    @abc.abstractmethod
    def connect(self):
        raise NotImplemented("baseConn-> Connect Method is not implemented !!!")

    @abc.abstractmethod
    def close(self):
        pass

    @abc.abstractmethod
    def create(self, stt=None, objName=None, addIndex=None):
        pass

    @abc.abstractmethod
    def getStructure(self, **args):
        pass

    @abc.abstractmethod
    def isExists(self, **args):
        pass

    @abc.abstractmethod
    def preLoading (self):
        pass

    @abc.abstractmethod
    def extract(self, **args):
        pass

    @abc.abstractmethod
    def load(self, **args):
        pass

    @abc.abstractmethod
    def execMethod(self, method=None):
        pass

    @abc.abstractmethod
    def merge(self, mergeTable, mergeKeys=None, sourceTable=None):
        pass

    """ -----------------   GLOBAL METHODS -------------------------------------"""
    """ General method - implemented localy """
    def dataTransform(self, data, functionDict=None, execDict=None):
        regex = r"(\{.*?\})"
        if (functionDict and len(functionDict) > 0) or (execDict and len(execDict) > 0):
            for num, dataRow in enumerate(data):
                row = list(dataRow)
                for ind in functionDict:
                    newVal = row[ind]
                    for fn in functionDict[ind]:
                        newVal = fn.handler(newVal, ind)
                    row[ind] = newVal

                for ind in execDict:
                    newVal = execDict[ind]
                    matches = re.finditer(regex, execDict[ind], re.MULTILINE | re.DOTALL)
                    for matchNum, match in enumerate(matches):
                        for groupNum in range(0, len(match.groups())):
                            colNum = match.group(1).replace('{', '').replace('}', '')
                            colVal = row[int(colNum)]
                            colVal = uniocdeStr(colVal,decode=True ) if colVal else ''
                            newVal = replaceStr(sString=str(newVal), findStr=match.group(1), repStr=colVal, ignoreCase=False, addQuotes=None)
                    row[ind] = newVal
                data[num] = row

        ## ceOBDC - convert data to None
        if self.conn == eConn.SQLSERVER:
            for num, row in enumerate(data):
                data[num] = [i if i!='' else None for i in row]
        return data

    def test(self):
        if self.connect():
            p("TEST-> SUCCESS: %s, type: %s " % (self.connName, self.conn))
        else:
            p("TEST-> FAILED: %s, type: %s " % (self.connName, self.conn))

    """ GET DATA TYPE TREE - Return source data types"""
    def getDataTypeTree (self, dataType, dataTypeTree=None, ret=list([])):
        dataTypeTree = dataTypeTree if dataTypeTree else  copy.copy(self.DATA_TYPES)
        for k in dataTypeTree:
            k = str(k)
            if k.lower() == dataType.lower():
                ret.append(k)
                return ret
            if isinstance(dataTypeTree[k], dict):
                retDic = self.getDataTypeTree(dataType=dataType, dataTypeTree=dataTypeTree[k], ret=ret)
                if retDic:
                    ret.append(k)
                    return ret
            elif isinstance(dataTypeTree[k], list):
                lowerDataType = [x.lower() for x in dataTypeTree[k]]
                if dataType.lower() in lowerDataType:
                    ret.append(str(dataType))
                    ret.append(k)
                    return ret
            elif dataType == dataTypeTree[k] or str(dataType).lower() == str(dataTypeTree[k]).lower():
                ret.append(str(dataType))
                ret.append(k)
                return ret
            elif dataType == k or str(dataType).lower() == str(k).lower():
                ret.append(k)
                return ret
        return None

    """ SET DARA TYPE - Return last known data type in target """
    def setDataTypeTree(self, dataTypeTree, allDataTypes, ret=[]):
        if not dataTypeTree or len(dataTypeTree)==0:
            return ret

        if isinstance(allDataTypes, list):
            for k in allDataTypes:
                if k in dataTypeTree:
                    ret.append(k)
                    dataTypeTree.remove(k)
                    return ret
            return ret
        elif allDataTypes and allDataTypes in dataTypeTree:
            return ret.append(allDataTypes)
        elif isinstance(allDataTypes, dict):
            for k in allDataTypes:
                if k in dataTypeTree:
                    ret.append(k)
                    if allDataTypes[k] and eConn.dataType.B_DEFAULT in allDataTypes[k]:
                        ret.append(allDataTypes[k][eConn.dataType.B_DEFAULT])

                    dataTypeTree.remove(k)
                    return self.setDataTypeTree(dataTypeTree=dataTypeTree, allDataTypes=allDataTypes[k], ret=ret)
        return ret

    """ return default data types dictionary   """

    def setDefaults(self, defaultsDic):
        defValues = {}
        if defaultsDic and eConn.NONO in defaultsDic and defaultsDic[eConn.NONO]:
            defValues = defaultsDic[eConn.NONO]

        if defaultsDic and self.conn in defaultsDic and defaultsDic[self.conn]:
            defValues.update(defaultsDic[self.conn])

        elif defaultsDic and len(defValues) == 0:
            defValues = defaultsDic

        for k in self.baseDefaults:
            if k not in defValues:
                defValues[k] = self.baseDefaults[k]
        return defValues

    """ Set DataTypes tree --> Adding new children """
    def setDataTypes (self, connDataTypes, existsDataTypes=None):
        existsDataTypes = self.baseDataTypes.copy() if not existsDataTypes else existsDataTypes
        connDataTypes = connDataTypes[self.conn] if self.conn in connDataTypes else connDataTypes
        if connDataTypes and len(connDataTypes)>0:
            for k in existsDataTypes:
                if k in connDataTypes:
                    existsDataTypes[k] = copy.copy(connDataTypes[k])
                elif existsDataTypes[k] and isinstance(existsDataTypes[k], dict):
                    self.setDataTypes (connDataTypes=connDataTypes, existsDataTypes=existsDataTypes[k])
        return existsDataTypes

    """ 
        propKey - property key to serch in connPropDict (default) or in propDict 
        if connProp is exists - return connProp
        if connPropDic is Dictionary and connKey in - return connPropDic value
        if no connProp and connKey not found in connPropDic return defValue
     """
    def setProperties (self, propKey, propVal, propDict=None, propDef=None):
        propDict = propDict if propDict else self.connPropDict

        if propVal and isinstance(propVal, dict) and propKey in propVal and propVal[propKey] is not None:
            return propVal[propKey]
        elif propVal is not None and not isinstance(propVal, dict):
            return propVal
        elif propDict and propKey in propDict and propDict[propKey] is not None:
            return propDict[propKey]
        elif isinstance(propDef, dict) and propKey in propDef:
            return propDef[propKey]
        return propDef

    def decodeStrPython2Or3(self, sObj, un=True, decode="windows-1255"):
        pVersion = sys.version_info[0]

        if 3 == pVersion:
            return sObj
        else:
            if un:
                return unicode(sObj)
            elif decode:
                return str(sObj).decode(decode)




# Copyright (c) 2017-2019, BPMK LTD (BiSkilled) Tal Shany <tal.shany@biSkilled.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import re
import copy
from collections import OrderedDict

from dingDong.misc.enumsJson    import eJson
from dingDong.misc.logger import p
from dingDong.conn.baseConnectorManager   import mngConnectors as conn
from dingDong.misc.misc         import uniocdeStr
from dingDong.config import config


class ddManager (object):
    def __init__(self, node, connDict=None):
        self.stt            = None
        self.addSourceColumn= True
        self.addIndex       = None
        self.nodes          = None
        self.connDict       = connDict if connDict else config.CONN_URL

        jsonNodes           = []

        if isinstance(node, (list,tuple)):
            jsonNodes  = node
        elif isinstance(node, (dict,OrderedDict)):
            jsonNodes = [node]
        else:
            p("NODE IS NOT LIST OR DICTIONARY, IGNORE NODE")

        ## INIT LIST NODES TO EXECUTE
        if len(jsonNodes)>0:
            self.nodes =  self.initNodes (jsonNodes)

    def initNodes (self, nodes):
        orderedNodes = []
        for node in nodes:
            modelDict = OrderedDict()
            if isinstance(node, (list, tuple)):
                orderedNodes.append (self.initNodes (nodes=node) )
            elif not isinstance(node, (dict, OrderedDict)):
                p("NODE IS NOT LIST OR DICT, IGNORE NODE\n%s" %str(node),"e")
            else:
                # ADD STT ONLY
                if eJson.jKeys.STTONLY in node:
                    self.stt = node[eJson.jKeys.STTONLY]
                    self.addSourceColumn = False

                # ADD STT
                if eJson.jKeys.STT in node:
                    self.stt = node[eJson.jKeys.STT]
                    self.addSourceColumn = True

                    if eJson.jKeys.STTONLY in node:
                        self.addSourceColumn = False
                        for k in node[eJson.jKeys.STTONLY]:
                            if k not in self.stt:
                                self.stt[k] = node[eJson.jKeys.STTONLY][k]
                            else:
                                self.stt[k].update ( node[eJson.jKeys.STTONLY][k] )

                # ADD Index
                if eJson.jKeys.INDEX in node:
                    self.addIndex = node[eJson.jKeys.INDEX]

                for i,k in enumerate (node):
                    if eJson.jKeys.SOURCE == k or eJson.jKeys.SOURCE in node[k]:
                        node[k][eJson.jValues.IS_SOURCE] = True
                        modelDict[eJson.jKeys.SOURCE] = conn(connPropDic=node[k], connLoadProp=self.connDict)

                    elif eJson.jKeys.QUERY == k or eJson.jKeys.QUERY in node[k]:
                        if eJson.jKeys.SOURCE in modelDict:
                            p("THERE IS QUERY AND SOURCE, WILL USE QUERY AS SOURCE", "w")
                        node[k][eJson.jValues.IS_SOURCE] = True
                        modelDict[eJson.jKeys.SOURCE] = conn(connPropDic=node[k], connLoadProp=self.connDict)

                    elif eJson.jKeys.TARGET == k or eJson.jKeys.TARGET in node[k]:
                        node[k][eJson.jValues.IS_TARGET] = True
                        modelDict[eJson.jKeys.TARGET] = conn(connPropDic=node[k], connLoadProp=self.connDict)

                    elif eJson.jKeys.MERGE == k or eJson.jKeys.MERGE in node[k]:
                        modelDict[eJson.jKeys.MERGE] = node[k]

                    elif eJson.jKeys.STT == k or eJson.jKeys.STTONLY==k or eJson.jKeys.INDEX==k:
                        pass
                    else:
                        modelDict[i] = conn(connPropDic=node[k], connLoadProp=self.connDict)
                orderedNodes.append(modelDict)

        return orderedNodes

    """ LOADING MODULE: Return mapping between source and target """
    def mappingLoadingSourceToTarget(self, src, tar):
        tarToSrc    = OrderedDict()
        srcStructure= src.getStructure()
        tarStructure= tar.getStructure()

        srcPre, srcPos, tarPre, tarPos = '', '', '', ''
        if hasattr(src, 'columnFrame'):
            srcPre, srcPos = src.columnFrame[0], src.columnFrame[1]

        if hasattr(tar, 'columnFrame'):
            tarPre, tarPos = tar.columnFrame[0], tar.columnFrame[1]

        # remove from STT column that not exists in Target OBJECT OR Source OBJECT

        self.__updateSTTBySourceOrTarget(srcStructure=srcStructure, pre=srcPre, pos=srcPos)
        srcColumns = OrderedDict()
        tarColumns = OrderedDict({x.replace(tarPre, "").replace(tarPos, "").lower(): x for x in tarStructure})

        sttColumns = OrderedDict({x.replace(tarPre, "").replace(tarPos, "").lower(): x for x in self.stt}) if self.stt else OrderedDict()

        ## {srcName in Target: Source column }
        for col in srcStructure:
            colAlias= col.replace(srcPre, "").replace(srcPos, "").lower()
            colName = srcStructure[col][eJson.jSttValues.SOURCE] if eJson.jSttValues.SOURCE in srcStructure[col] else col
            srcColumns[colAlias] = colName

        # There is no target schema --> using all source and STT
        if tar.usingSchema and self.addSourceColumn:
            for col in srcColumns:
                tarToSrc[col] = {eJson.jSttValues.SOURCE: srcColumns[col]}

        # ADD ALL COLUMN FROM SOURCE
        elif self.addSourceColumn:
            for col in tarColumns:
                if col in srcColumns:
                    tarToSrc[tarColumns[col]] = {eJson.jSttValues.SOURCE: srcColumns[col]}

        tarToSrcColumns = {x.replace(tarPre, "").replace(tarPos, "").lower(): x for x in tarToSrc}

        for col in sttColumns:
            if col in tarToSrcColumns:
                tarToSrc[tarToSrcColumns[col]].update(self.stt[sttColumns[col]])
            else:
                tarToSrc[sttColumns[col]] = self.stt[sttColumns[col]]

        if not tar.usingSchema:
            p("TAREGT %s DO NOT HAVE FIX SCHEMA " %(tar.conn), "ii")
            return tarToSrc

        #### Check Column in Source and not exists in mapping
        existsTarColumns = {}
        existsSrcColumns = {}

        for col in tarToSrc:
            colL = col.replace(tarPre, "").replace(tarPos, "").lower()
            existsTarColumns[colL] = col
            if eJson.jSttValues.SOURCE in tarToSrc[col] and tarToSrc[col][eJson.jSttValues.SOURCE]:
                srcL = tarToSrc[col][eJson.jSttValues.SOURCE].replace(srcPre, "").replace(srcPos, "").lower()
                existsSrcColumns[srcL] = tarToSrc[col][eJson.jSttValues.SOURCE]

        columnNotMapped = u""
        for col in tarColumns:
            if col not in existsTarColumns:
                columnNotMapped+=uniocdeStr(tarColumns[col])+u" ; "
        if len(columnNotMapped) > 0:
            p(u"TARGET COLUMN NOT MAPPED: %s" % (columnNotMapped), "w")

        columnNotMapped = u""
        for col in srcColumns:
            if srcColumns[col].replace(srcPre, "").replace(srcPos, "").lower() not in existsSrcColumns:
                columnNotMapped+=uniocdeStr(srcColumns[col])+u" ; "
        if len(columnNotMapped) > 0:
            p(u"SOURCE COLUMN NOT MAPPED: %s" % (columnNotMapped), "w")

        return tarToSrc

    """ MAPPING MODULE: Convert Source Data Type into Target Data Type """
    def updateTargetBySourceAndStt(self, src, tar):
        retStrucure = OrderedDict()
        sourceStt   = src.getStructure()
        sourceColL  = {x.lower():x for x in sourceStt}

        if src.conn == tar.conn:
            if self.addSourceColumn:
                for col in sourceStt:
                    if eJson.jSttValues.ALIACE in sourceStt[col] and sourceStt[col][eJson.jSttValues.ALIACE]:
                        retStrucure[ sourceStt[col][eJson.jSttValues.ALIACE] ] = {eJson.jSttValues.TYPE: sourceStt[col][eJson.jSttValues.TYPE]}
                    else:
                        retStrucure[col] = {eJson.jSttValues.TYPE: sourceStt[col][eJson.jSttValues.TYPE]}

        ### Source connection type is different than target connection type
        elif self.addSourceColumn:

            for col in sourceStt:
                targetColName = col
                if eJson.jSttValues.ALIACE in sourceStt[col] and sourceStt[col][eJson.jSttValues.ALIACE]:
                    targetColName = sourceStt[col][eJson.jSttValues.ALIACE]

                colType = sourceStt[col][eJson.jSttValues.TYPE] if eJson.jSttValues.TYPE in sourceStt[col] and sourceStt[col][eJson.jSttValues.TYPE] else tar.defDataType
                fmatch = re.search(r'(.*)(\(.+\))', colType, re.M | re.I)
                if fmatch:
                    replaceString   = fmatch.group(1)  # --> varchar, int , ...
                    postType        = fmatch.group(2)  # --> (X), (X,X) , ....
                else:
                    replaceString   = colType
                    postType        = ''

                ## Receive list of all dataType in DataTypes Tree
                newDataTypeTree     = src.getDataTypeTree (dataType=replaceString.lower(), ret=([]))
                if newDataTypeTree is None:
                    p("SOURCE CONNECTION: %s, COLUMN: %s, DATA TYPE: %s ; IS NOT EXISTS, WILL USE DEFAULT VALUE" %(src.conn,col, replaceString),"w")
                    tarType = tar.defDataType
                else:
                    targetList = tar.setDataTypeTree (dataTypeTree=newDataTypeTree, allDataTypes=tar.DATA_TYPES, ret=[])
                    tarType = '%s%s' %(targetList[-1],postType) if targetList and len(targetList)>0 else tar.defDatatType
                retStrucure[targetColName] = {eJson.jSttValues.TYPE:tarType}

        retStrucureL = {x.lower():x for x in retStrucure}
        retStrucureKeys = list (retStrucure.keys())

        if self.stt:
            for col in self.stt:
                if eJson.jSttValues.SOURCE in self.stt[col]:
                    sourceName = self.stt[col][eJson.jSttValues.SOURCE]
                    if sourceName in retStrucureKeys:
                        del retStrucure[sourceName]
                    elif sourceName.lower() in retStrucureL:
                        del retStrucure[retStrucureL[sourceName.lower()]]

                if col.lower() in sourceColL and eJson.jSttValues.TYPE in self.stt[col] and self.stt[col][eJson.jSttValues.TYPE]:
                    if sourceColL[ col.lower() ] not in retStrucure:
                        retStrucure[ sourceColL[ col.lower() ] ] = {}
                    retStrucure[sourceColL[ col.lower() ]] [eJson.jSttValues.TYPE] = self.stt[col][eJson.jSttValues.TYPE]

                    if eJson.jSttValues.ALIACE in self.stt[col]:
                        retStrucure[sourceColL[col.lower()]][eJson.jSttValues.ALIACE] = self.stt[col][
                            eJson.jSttValues.ALIACE]

                elif col.lower() not in sourceColL:
                    if eJson.jSttValues.TYPE in self.stt[col] and self.stt[col][eJson.jSttValues.TYPE]:
                        retStrucure[col] = {eJson.jSttValues.TYPE: self.stt[col][eJson.jSttValues.TYPE]}
                    else:
                        retStrucure[col] = {eJson.jSttValues.TYPE: tar.defDataType}

                        if eJson.jSttValues.ALIACE in self.stt[col]:
                            retStrucure[col][eJson.jSttValues.ALIACE] = self.stt[col][eJson.jSttValues.ALIACE]

        return retStrucure

    """ Check Source values in STT - remove invalid values """
    def __updateSTTBySourceOrTarget (self, srcStructure, pre="[", pos="]"):
        # Check if ther are sourcea in STT that not defined
        srcStrucureL = []
        srcColumns = {}

        for col in  srcStructure:
            srcStrucureL.append (col.replace(pre,"").replace(pos,"").lower())
            if eJson.jSttValues.SOURCE in srcStructure[col] and srcStructure[col][eJson.jSttValues.SOURCE]:
                srcStrucureL.append(uniocdeStr(srcStructure[col][eJson.jSttValues.SOURCE]))
                srcColumns[srcStructure[col][eJson.jSttValues.SOURCE]] = None

        removeColumnsSrc = []
        if self.stt:
            for col in self.stt:
                if eJson.jSttValues.SOURCE in self.stt[col] and self.stt[col][eJson.jSttValues.SOURCE] not in srcColumns:
                    if self.stt[col][eJson.jSttValues.SOURCE].replace(pre,"").replace(pos,"").lower() not in srcStrucureL:
                        removeColumnsSrc.append (col)

            for col in removeColumnsSrc:
                p("STT TAREGT %s HAVE INVALID SOURCE %s --> ignore COLUMN " % (col, self.stt[col][eJson.jSttValues.SOURCE]),"w")
                del self.stt[col]

    def dong( self ):
        if self.nodes and len(self.nodes)>0:
            src         = None
            tar         = None
            mrg         = None
            mrgSource   = None
            for node in self.nodes:
                for k in node:
                    ## Exec METHOD
                    if str(k).isdigit():
                        node[k].execMethod()

                    if eJson.jKeys.SOURCE == k:
                        src = node[k]
                        mrgSource = src

                    elif eJson.jKeys.TARGET == k:
                        tar = node[k]
                        mrgSource = tar

                    elif eJson.jKeys.MERGE == k:
                        mrg = node[k]

                    if src and tar:
                        """ TRANSFER DATA FROM SOURCE TO TARGET """
                        tar.preLoading()
                        mrgSource = tar
                        tarToSrc = self.mappingLoadingSourceToTarget(src=src, tar=tar)
                        src.extract(tar=tar, tarToSrc=tarToSrc, addAsTaret=True)
                        tar.close()
                        src.close()
                        src = None

                    """ MERGE DATA BETWEEN SOURCE AND TARGET TABLES """
                    if mrg and mrgSource:
                        mrgSource.connect()
                        mergeTarget = mrg[eJson.jMergeValues.TARGET]
                        mergeKeys   = mrg[eJson.jMergeValues.MERGE]
                        mrgSource.merge(mergeTable=mergeTarget, mergeKeys=mergeKeys, sourceTable=None)
                        mrgSource.close()

    def ding( self ):
        if self.nodes and len(self.nodes) > 0:
            src         = None
            tar         = None
            mrgSource   = None
            for node in self.nodes:
                for k in node:
                    if eJson.jKeys.SOURCE == k:
                        src = node[k]
                        mrgSource = src

                    elif eJson.jKeys.TARGET == k:
                        tar = node[k]
                        mrgSource = tar

                        if  eJson.jKeys.SOURCE not in node:
                            tar.create(stt=self.stt, addIndex=self.addIndex)
                            tar.close()
                            tar = None

                    if tar and src:
                        # convert source data type to target data types
                        targetStt = self.updateTargetBySourceAndStt(src=src, tar=tar)
                        tar.create(stt=targetStt, addIndex=self.addIndex)
                        mrgSource = tar
                        src.close()
                        tar.close()
                        src = None

                    if eJson.jKeys.MERGE == k and mrgSource:
                        mrgTarget = copy.copy(mrgSource)
                        mrgSource.connect()
                        mrgTarget.connect()
                        mrgTarget.connObj   = node[k][eJson.jMergeValues.TARGET]
                        mrgTarget.connIsTar = True
                        mrgTarget.connIsSrc = False
                        sttMerge = self.updateTargetBySourceAndStt(src=mrgSource, tar=mrgTarget)
                        mrgTarget.create(stt=sttMerge,addIndex=self.addIndex)
                        mrgTarget.close()
                        mrgSource.close()
                        mrgSource = None


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.


# create HTML

class eHtml (object):
    HEADER  = "header"
    ROWS    = "rows"

def createHtmlFromList(htmlList, htmlHeader=None):
    htmlRow = "\t\t<h1>%s</h1>\n" %htmlHeader if htmlHeader else ""
    ### color : #000000
    htmlHeader = """
            table {
                border-collapse: collapse; 
                border:1px solid #69899F;}
            table td{
                border:1px dotted black;
                padding:5px;}
            table td:first-child{
                border-left:0px solid black;}
            table th{
               border:2px solid #69899F;
               padding:5px;}"""

    htmlHeader = "\t<style>%s\n\t\t</style>" %htmlHeader

    htmlHeader = """
    <header>
    %s
    </header>""" %(htmlHeader)

    htmlBody = ""
    for tblDic in htmlList:
        lHeader     = None
        lRows       = None
        tblHeader   = ""
        tblLines    = ""

        if eHtml.HEADER in tblDic:
            lHeader = tblDic[eHtml.HEADER]
            for h in lHeader:
                tblHeader += '<th>%s</th>' % str(h)

            tblHeader = """
            <thead>
                <tr>
                %s
                </tr>
            </thead>""" %tblHeader

        if eHtml.ROWS in tblDic:
            lRows   = tblDic[eHtml.ROWS]
            for row in lRows:
                tblLinesTD = ""
                for col in row:
                    strRow = ""
                    if isinstance(col, (list,tuple)):
                        for c in col:
                            strRow+= "%s <br>" %str(c)
                    else:
                        strRow = str(col)
                    tblLinesTD += '<td>%s</td>' %strRow
                tblLines+=  "<tr>\n\t\t\t\t\t%s\n\t\t\t\t</tr>\n\t\t\t\t" %tblLinesTD
            tblLines = """
            <tbody class="allign-center">
                %s
            </tbody>    
            """ %tblLines

        tblHtml = '\t\t<table class="table table-bordered">%s%s\n\t\t</table>' %(tblHeader, tblLines)

        htmlBody+='%s\n\t<br>\n' %tblHtml

    htmlBody = "\t<body dir='ltr'>\n%s%s</body>" %(htmlRow,htmlBody)

    htmlStr = "<html>%s\n%s\n</html>" %(htmlHeader, htmlBody)

    return htmlStr

#htmlList = [{'header':['col1','col2','col3','col4'],'rows':[['a1','a2','a3','a4'],['b1','b2','b3','b4'],['c1','c2','c3','c4']]},
#            {'header':['col1','col2','col3','col4'],'rows':[['a1','a2','a3','a4'],['b1','b2','b3','b4'],['c1','c2','c3','c4']]}]
#xx = createHtmlFromList(htmlList, "hallo world")
#print (xx)







# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

from dingDong.misc.logger import  p
from dingDong.misc.misc import uniocdeStr


def OLAP_Process(serverName,dbName, cubes=[], dims=[], fullProcess=True):
    import sys, os
    localPath = os.path.abspath(os.path.dirname(__file__))
    sys.path.append(os.path.join(localPath, r'../dll/clrmodule.dll"'))
    import clr
    clr.AddReference(os.path.join(localPath, r'../dll/Microsoft.AnalysisServices.DLL') )

    from Microsoft.AnalysisServices import Server
    from Microsoft.AnalysisServices import ProcessType

    processType = ProcessType.ProcessFull if fullProcess else 0
    # Connect to server
    amoServer = Server()
    amoServer.Connect(serverName)

    # Connect to database
    amoDb = amoServer.Databases[dbName]

    for dim in amoDb.Dimensions:
        if len(dims)==0 or dim in dims:
            try:
                dim.Process(processType)
                p(u"OLAP DB: %s, process DIM %s finish succeffully ... " %(uniocdeStr(dbName, decode=True),uniocdeStr(dim, decode=True)), "i")
            except Exception as e:
                p(u"OLAP DB: %s, ERROR processing DIM %s ... " % (uniocdeStr(dbName, decode=True),uniocdeStr(dim, decode=True)),"e")
                p(e,"e")

    for cube in amoDb.Cubes:
        if len(cubes)==0 or cube in cubes:
            try:
                cube.Process(processType)
                p(u"OLAP DB: %s, CUBE %s finish succeffully ... " %(uniocdeStr(dbName, decode=True),uniocdeStr(cube, decode=True)),"i")
            except Exception as e:
                p(u"OLAP DB: %s, ERROR processing CUBE %s ... " % (uniocdeStr(dbName, decode=True), uniocdeStr(cube, decode=True)),"e")
                p(e,"e")


# Copyright (c) 2017-2019, BPMK LTD (BiSkilled) Tal Shany <tal.shany@biSkilled.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import os
import io
import json

from collections import OrderedDict

from dingDong.misc.logger import p
from dingDong.misc.enumsJson import eJson, eConn, findProp
from dingDong.config      import config


#xx = [[{'s':}],[{'prop':{'pp':45}},{}],{'sql1':'', 's':xxx,'t':'dsdsdsd'},{''},{}]
# s : [obj] [conn,  obj],[conn. obj, filter]  {"conn":conn, "obj":obj, filter:"dddd"} // connection:"s":[]

class jsonParser (object):

    """ Create list of node to load and init all connection dictionary """
    def __init__ (self, dicObj=None, filePath=None,
                  dirData=None, includeFiles=None, notIncludeFiles=None, connDict=None, sqlFolder=None):

        if connDict and isinstance(connDict ,(dict, OrderedDict)):
            config.CONN_URL.update(connDict)

        self.connDict = config.CONN_URL
        self.sqlFolder= sqlFolder
        self.__initConnDict ()
        self.listObj    = []
        self.listFiles  = []
        self.jsonMapp   = []
        self.jsonName   = ''
        self.dirData    = None
        msg     = ""

        if dicObj:
            self.listObj.append (dicObj)
            msg+="loading data from dictioanry; "

        if filePath:
            if os.path.isfile (filePath):
                with io.open(filePath, encoding="utf-8") as jsonFile:  #
                    dicObj = json.load(jsonFile, object_pairs_hook=OrderedDict)
                    self.listObj.append (dicObj)
                    msg+="loading data from file"
            else:
                p("file %s is not exists " %(filePath),"e")

        if dirData:
            self.dirData = dirData
            if not os.path.isdir(dirData):
                p ("Folder not  exists : %s" % dirData, "e")
                return
            else:
                msg += "loading data from folder: %s" %dirData

                jsonFiles = [pos_json for pos_json in os.listdir(dirData) if pos_json.endswith('.json')]
                jsonFilesDic = {x.lower().replace(".json", ""): x for x in jsonFiles}

                if notIncludeFiles:
                    notIncludeDict = {x.lower().replace(".json", ""): x for x in notIncludeFiles}
                    for f in jsonFilesDic:
                        if f in notIncludeDict:
                            p('NOT INCLUDE: File:%s, file: NOT IN USED, REMOVED >>>>' % (str( os.path.join(dirData,f) ) ), "ii")
                            jsonFiles.remove(jsonFilesDic[f])
                    for f in notIncludeDict:
                        if f not in jsonFilesDic:
                            p('NOT INCLUDE: File: %s, NOT EXISTS.Ignoring>>>>>' % ( str(os.path.join(dirData, f))), "ii")

                if includeFiles:
                    includeDict = {x.lower().replace(".json", ""): x for x in includeFiles}
                    for f in jsonFilesDic:
                        if f not in includeDict and f in jsonFilesDic and jsonFilesDic[f] in jsonFiles:
                            p('INCLUDE: Folder:%s, file: %s NOT IN USED, REMOVED >>>>' % (str(dirData), f), "ii")
                            jsonFiles.remove(jsonFilesDic[f])
                    for f in includeDict:
                        if f not in jsonFilesDic:
                            p('INCLUDE: Folder: %s, file: %s not exists.. Ignoring >>>>' % (str(dirData), f), "ii")
                self.listFiles = jsonFiles

    def jsonMappings (self, destList = None):
        if self.listObj and len (self.listObj)>0:
            self.jsonMapp = list([])
            self.jsonName = ''
            self.__initMetaDict(listObj=self.listObj, destList=None)

            yield {self.jsonName:self.jsonMapp}
        elif self.listFiles and len (self.listFiles)>0:
            for index, js in enumerate(self.listFiles):
                self.jsonMapp = list([])
                self.jsonName = js
                with io.open(os.path.join(self.dirData, js), encoding="utf-8") as jsonFile:  #
                    jText = json.load(jsonFile, object_pairs_hook=OrderedDict)
                    self.__initMetaDict(listObj=jText, destList=None)
                    yield {self.jsonName:self.jsonMapp}

    def getAllConnection (self, pr=True):
        if pr:
            for x in self.connDict:
                p('TYPE: %s\t\t\tPROP: %s' %(str(x), str(self.connDict[x])), "ii")
            p ("====================================================" , "ii")
        return self.connDict

    """ MAIN KEYS : NAME (n) AND CONNENCTION TYPE (conn) """
    def __initConnDict (self):
        newConnDict =  {}
        for conn in self.connDict:
            dictProp = eJson.jValues.DIC.copy()
            dictProp[eJson.jValues.NAME] = conn
            dictProp[eJson.jValues.CONN] = conn

            if isinstance( self.connDict[conn], dict ):
                for k in self.connDict[conn]:
                    origK = findProp(prop=k, obj=eJson.jValues )
                    if origK:
                        dictProp[origK] = self.connDict[conn][k]
                    elif k.lower() in newConnDict:
                        dictProp[k.lower()] = self.connDict[conn][k]
                    else:
                        dictProp[k] = self.connDict[conn][k]

            elif isinstance( self.connDict[conn], str ):
                dictProp[eJson.jValues.URL] = self.connDict[conn]
            newConnDict[conn] = dictProp

            ### VALID CONNECTION TYPES
            errConn = []
            for conn in newConnDict:
                if not findProp(prop=newConnDict[conn][eJson.jValues.CONN], obj=eConn):
                    p("Remove Connection %s becouse prop: %s  is NOT VALID !!" %(str(conn), str(newConnDict[conn][eJson.jValues.CONN])), "e")
                    errConn.append (conn)
            for err in errConn:
                del newConnDict[err]
        self.connDict = newConnDict

    # parse Json file into internal Json format
    # Key can be name or internal dictionary
    def __initMetaDict (self, listObj,destList=None):
        for node in listObj:
            if isinstance(node ,list):
                self.__initMetaDict(listObj=node, destList=destList)
            elif isinstance(node, dict):
                stt     = OrderedDict()
                newDict = OrderedDict()

                for prop in node:
                    k =  findProp (prop=prop.lower(), obj=eJson.jKeys, dictProp=node[prop])
                    if k:
                        if eJson.jKeys.SOURCE  in newDict and k == eJson.jKeys.QUERY:
                            p("Source and Query exists - Will use Query", "ii")
                            del newDict[eJson.jKeys.SOURCE]
                        elif eJson.jKeys.QUERY in newDict and k == eJson.jKeys.SOURCE:
                            p("Query and Source exists - Will use Source", "ii")
                            del newDict[eJson.jKeys.QUERY]

                        if k == eJson.jKeys.STT or k == eJson.jKeys.STTONLY:
                            newSttL = {x.lower(): x for x in node[prop]}
                            sttL = {x.lower(): x for x in stt}
                            for s in stt:
                                if s.lower() in newSttL:
                                    stt[s].update(node[prop][newSttL[s.lower()]])

                            for s in newSttL:
                                if s not in sttL:
                                    stt[newSttL[s]] = node[prop][newSttL[s]]
                            newDict[k] =stt
                        # parse source / target / query
                        elif k == eJson.jKeys.SOURCE or k == eJson.jKeys.TARGET or k == eJson.jKeys.QUERY:
                            newDict[k] = self.__sourceOrTargetOrQueryConn(propFullName=prop, propVal=node[prop])
                        # parse merge
                        elif k == eJson.jKeys.MERGE:
                            newDict[k] = self.__mergeConn(existsNodes=newDict, propFullName=prop, propVal=node[prop])
                        # parse column data types
                        elif k == eJson.jKeys.COLUMNS:
                            stt = self.__sttAddColumns (stt=stt, propVal=node[prop])
                            newDict[eJson.jKeys.STTONLY] = stt

                            if eJson.jKeys.STT in newDict:
                                del newDict[eJson.jKeys.STT]

                        # parse column mapping
                        elif k == eJson.jKeys.MAP:
                            stt = self.__sttAddMappings(stt=stt, propVal=node[prop])
                            if eJson.jKeys.STT in newDict:
                                newDict[eJson.jKeys.STT] = stt
                            else:
                                newDict[eJson.jKeys.STTONLY] = stt
                        elif k == eJson.jKeys.INDEX:
                            index = self.__index(propVal=node[prop])
                            if index:
                                newDict[eJson.jKeys.INDEX] = index
                        else:
                            p ("%s not implemented !" %(k), "e")
                    else:
                        newDict[prop.lower()] = self.__uniqueProc(propVal=node[prop])

                self.jsonMapp.append ( newDict )
            else:
                p("jsonParser->__initMetaDict: Unknown node prop values: %s " %node)

    ### List option : [obj] [conn,  obj],[conn. obj, filter]  // connection Name defoult
    ### FINAL : {eJK.CONN:None, eJK.OBJ:None, eJK.FILTER:None, eJK.URL:None, eJK.URLPARAM:None}
    def __sourceOrTargetOrQueryConn(self, propFullName, propVal):
        ret = eJson.jValues.DIC.copy()
        if isinstance(propVal, str):
            ret[eJson.jValues.NAME] = propFullName
            ret[eJson.jValues.TYPE] = propFullName
            ret[eJson.jValues.OBJ]  = propVal
        elif isinstance(propVal, list):
            ret[eJson.jValues.NAME] = propFullName
            ret[eJson.jValues.TYPE] = propFullName
            if len(propVal) == 1:
                ret[eJson.jValues.CONN] = propFullName
                ret[eJson.jValues.OBJ] = propVal[0]
            elif len(propVal) == 2:
                ret[eJson.jValues.CONN] = propVal[0]
                ret[eJson.jValues.OBJ]  = propVal[1]
            elif len(propVal) == 3:
                ret[eJson.jValues.CONN]  = propVal[0]
                ret[eJson.jValues.OBJ]   = propVal[1]

                if str(propVal[2]).isdigit():
                    ret[eJson.jValues.UPDATE] =  self.__setUpdate (propVal[2])
                else:
                    ret[eJson.jValues.FILTER]= propVal[2]
            elif len(propVal) == 4:
                ret[eJson.jValues.CONN]     = propVal[0]
                ret[eJson.jValues.OBJ]      = propVal[1]
                ret[eJson.jValues.FILTER]   = propVal[2]
                ret[eJson.jValues.UPDATE]   = self.__setUpdate (propVal[3])

            else:
                p("%s: Not valid list valuues, must 1,2 or 3" % (str(propFullName)), "e")
        elif isinstance(propVal, dict):
            self.__notVaildProp(ret, propVal, propFullName)
        else:
            p("Not valid values: %s " %(propVal),"e")
            return {}

        if findProp (prop=ret[eJson.jValues.TYPE], obj=eJson.jKeys)  == eJson.jKeys.QUERY:
            ret[eJson.jValues.IS_SQL] = True
        return ret

    # Must have source / target / query in existing nodes
    # [obj], [obj, keys]
    def __mergeConn (self, existsNodes, propFullName, propVal):
        ret     = eJson.jMergeValues.DIC.copy()
        srcConn = None

        if eJson.jKeys.TARGET in existsNodes:
            srcConn = existsNodes[eJson.jKeys.TARGET]
        elif eJson.jKeys.SOURCE in existsNodes:
            srcConn = existsNodes[eJson.jKeys.SOURCE]
        elif eJson.jKeys.QUERY in existsNodes:
            srcConn = existsNodes[eJson.jKeys.QUERY]

        if not srcConn:
            p("There is no query/ source or target to merge with. quiting. val: %s " %(str(propVal)) ,"e")
            return {}

        ### Update values from Source connection
        ret[eJson.jMergeValues.SOURCE]  = srcConn[eJson.jValues.OBJ]
        ret[eJson.jValues.CONN]         = srcConn[eJson.jValues.CONN]
        ret[eJson.jValues.URL]          = srcConn[eJson.jValues.URL]
        ret[eJson.jValues.URLPARAM]     = srcConn[eJson.jValues.URLPARAM]
        ret[eJson.jValues.URL_FILE]     = srcConn[eJson.jValues.URL_FILE]

        if isinstance(propVal, str):
            ret[eJson.jValues.NAME]         = propFullName
            ret[eJson.jMergeValues.TARGET]  = propVal

        elif isinstance(propVal, list):
            ret[eJson.jValues.NAME] = propFullName
            if len(propVal) == 1:
                ret[eJson.jMergeValues.TARGET]  = propVal[0]
            elif len(propVal) == 2:
                ret[eJson.jMergeValues.TARGET] = propVal[0]
                if str(propVal[1]).isdigit():
                    ret[eJson.jValues.UPDATE] = self.__setUpdate(propVal[1])
                else:
                    ret[eJson.jMergeValues.MERGE]  = propVal[1]
            elif len(propVal) == 3:
                ret[eJson.jMergeValues.TARGET]  = propVal[0]
                ret[eJson.jMergeValues.MERGE]   = propVal[1]
                ret[eJson.jValues.UPDATE]       = self.__setUpdate (propVal[2])
            else:
                p("%s: Not valid merge valuues, must have obj and merge key..." % (str(propVal)), "e")
        elif isinstance(propVal, dict):
            self.__notVaildProp( ret, propVal, propFullName )
        else:
            p("Not valid values")

        ret[eJson.jValues.OBJ] = ret[eJson.jMergeValues.TARGET]

        return ret

    def __setUpdate (self, val):
        if str(val).isdigit():
            if findProp(prop=val, obj=eJson.jUpdate):
                return val
            else:
                p("THERE IS %s WHICH IS MAPPED TO UPDATE PROPERTY, MUST HAVE -1(drop), 1(UPDATE), 2(NO_UPDATE), USING -1 DROP--> CREATE METHOD ")
        return -1

    # Insert into STT Column mapping
    def __sttAddColumns(self, stt, propVal):
        if not isinstance(propVal, dict):
            p ("jsonParser->__sttAddColumns: Not valid prop %s, must be dictionary type" %(propVal),"e")
            return stt

        existsColumnsDict   = {x.lower():x for x in stt.keys()}


        for tar in propVal:
            if tar.lower() in existsColumnsDict:
                stt[ existsColumnsDict[tar.lower()] ][eJson.jSttValues.TYPE] = propVal[tar]
            else:
                stt[tar] = {eJson.jSttValues.TYPE:propVal[tar]}
        return stt

    # Insert / Add new column types
    def __sttAddMappings(selfself, stt, propVal):
        if not isinstance(propVal, dict):
            p ("jsonParser->__sttAddMappings: Not valid prop %s, must be dictionary type" %(propVal),"e")
            return stt
        existsColumnsDict = {x.lower(): x for x in stt.keys()}
        for tar in propVal:
            if tar.lower() in existsColumnsDict:
                stt[ existsColumnsDict[tar.lower()] ][eJson.jSttValues.SOURCE] = propVal[tar]
            else:
                stt[tar][eJson.jSttValues.SOURCE] = propVal[tar]
        return stt

    # Special connection execution
    # list - ]
    def __uniqueProc(self, propVal):
        ret= {}
        if isinstance(propVal ,list) and len(propVal)==2:
            ret[eJson.jValues.CONN]     = propVal[0]
            ret[eJson.jValues.OBJ]      = propVal[1]
            ret[eJson.jValues.FOLDER]   = self.sqlFolder

        elif isinstance(propVal ,dict) and eJson.jValues.CONN in propVal and eJson.jValues.OBJ in propVal:
            ret = propVal

        if eJson.jValues.OBJ in ret and ret[eJson.jValues.OBJ] is not None and '.sql' in ret[eJson.jValues.OBJ]:
            fileName = ret[eJson.jValues.OBJ]
            if os.path.isfile(fileName):
                ret[eJson.jValues.FILE] = fileName
            if eJson.jValues.FOLDER in ret and ret[eJson.jValues.FOLDER] is not None:
                folderPath = ret[eJson.jValues.FOLDER]
                if os.path.isfile( os.path.join(folderPath, fileName)):
                    ret[eJson.jValues.FILE] = os.path.join(folderPath, fileName)

        return ret

    def __index (self, propVal):
        ## propVal = [{col:[],'iu':True, 'ic':True},{}]
        if isinstance(propVal, (dict,OrderedDict)):
            propVal = [propVal]

        if not isinstance(propVal, list):
            p("INDEX VALUES MUST BE A DICTIONARY OR LIST OF DICTIOANRY, FORMAT {'C'':list_column_index, 'ic':is cluster (True/False), 'iu': is unique (True/False)}")
            return
        ret = []
        for indexDict in propVal:

            if not isinstance(indexDict, (dict, OrderedDict)):
                p("INDEX MUST BE DICTIOANY, FORMAT {'C'':list_column_index, 'ic':is cluster (True/False), 'iu': is unique (True/False)}")
                continue

            returnDict = {eJson.jValues.INDEX_COLUMS:[],eJson.jValues.INDEX_CLUSTER:True,eJson.jValues.INDEX_UNIQUE:False}
            for node in indexDict:
                k =  findProp (prop=node.lower(), obj=eJson.jValues, dictProp=indexDict[node])

                if not k:
                    p("INDEX VALUES IS NOT VALID: %s, IGNORE INDEX. VALID FORMAT: FORMAT {'C'':list_column_index, 'ic':is cluster (True/False), 'iu': is unique (True/False)}")
                    break

                if k == eJson.jValues.INDEX_COLUMS:
                    if isinstance(indexDict[node], list ):
                        returnDict[eJson.jValues.INDEX_COLUMS].extend(indexDict[node])
                    else:
                        returnDict[eJson.jValues.INDEX_COLUMS].append(indexDict[node])
                elif k == eJson.jValues.INDEX_CLUSTER:
                    if not indexDict[node]: returnDict[eJson.jValues.INDEX_CLUSTER] = False
                elif k == eJson.jValues.INDEX_UNIQUE:
                    if indexDict[node]: returnDict[eJson.jValues.INDEX_UNIQUE] = True
                else:
                    p("INDEX - UNRECOGNIZED KEY %s IN DICT:%s IGNORE. VALID FORMAT: FORMAT {'C'':list_column_index, 'ic':is cluster (True/False), 'iu': is unique (True/False)}" %(str(node),str(indexDict)), "e")
            ret.append(indexDict)
        return ret


    def __notVaildProp(self, propDic, newPropDic, propFullName):
        for k in newPropDic:
            if k not in propDic:
                p("baseParser->notValidProp: %s: not valid prop %s -> ignore" % (propFullName, k), "e")
                del newPropDic[k]
        return newPropDic


# -*- coding: utf-8 -*-
# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import os
import io
import sys
import time
import traceback
from collections import OrderedDict

from dingDong.conn.baseBatch            import baseBatch
from dingDong.conn.baseBatchFunction    import *
from dingDong.misc.enumsJson            import eConn, eJson, eSql
from dingDong.misc.misc                 import uniocdeStr
from dingDong.config                    import config
from dingDong.misc.logger               import p
import dingDong.conn.connGlobalDbQueryParser as qp
from dingDong.conn.connGlobalDbSqlQueries import setSqlQuery
from dingDong.executers.executeSql import execQuery

try:
    import cx_Oracle  # version : 6.1
except ImportError:
    p("cx_Oracle is not installed", "ii")

DEFAULTS = {
            eConn.NONO: {   eJson.jValues.DEFAULT_TYPE:'varchar(100)',eJson.jValues.SCHEMA:'dbo',
                            eJson.jValues.EMPTY:'Null',eJson.jValues.COLFRAME:("[","]"),
                            eJson.jValues.SP:{'match':None, 'replace':None}},

            eConn.ORACLE: {eJson.jValues.DEFAULT_TYPE: 'varchar(100)', eJson.jValues.SCHEMA: 'dbo',
                           eJson.jValues.EMPTY: 'Null', eJson.jValues.COLFRAME: ('"', '"'),
                           eJson.jValues.SP: {'match': r'([@].*[=])(.*?(;|$))', 'replace': r"[=;@\s']"}},

            eConn.SQLSERVER: {eJson.jValues.DEFAULT_TYPE: 'varchar(100)', eJson.jValues.SCHEMA: 'dbo',
                              eJson.jValues.EMPTY: 'Null', eJson.jValues.COLFRAME: ("[", "]"),
                              eJson.jValues.SP: {'match': r'([@].*[=])(.*?(;|$))', 'replace': r"[=;@\s']"}  },

            eConn.LITE: {   eJson.jValues.DEFAULT_TYPE:'varchar(100)',eJson.jValues.SCHEMA:None,
                            eJson.jValues.EMPTY:'Null'}
           }

DATA_TYPES = {
    eConn.ORACLE:   { eConn.dataType.DB_DATE:['date','datetime'],
                      eConn.dataType.DB_VARCHAR:['varchar','varchar2'],
                      eConn.dataType.DB_DECIMAL:['number','numeric','dec','decimal']
                    },
    eConn.SQLSERVER:{
                        eConn.dataType.DB_DATE:['smalldatetime','datetime'],
                        eConn.dataType.DB_DECIMAL:['decimal']
                    },

    eConn.ACCESS: { eConn.dataType.DB_VARCHAR:['varchar', 'longchar', 'bit', 'ntext'],
                    eConn.dataType.DB_INT:['integer', 'counter'],
                    eConn.dataType.DB_FLOAT:['double'],
                    eConn.dataType.DB_DECIMAL:['decimal']
                    }
}

class baseGlobalDb (baseBatch):

    def __init__ (self, connPropDict=None, conn=None, connUrl=None, connExtraUrl=None,
                  connName=None,connObj=None,  connFilter=None, connIsTar=None,
                  connIsSrc=None, connIsSql=None):
        baseBatch.__init__(self, conn=conn, connName=connName, connPropDict=connPropDict, defaults=DEFAULTS, dataType=DATA_TYPES)
        self.usingSchema= True

        """ BASIC PROPERTIES FROM BASECONN """
        self.conn           = self.conn
        self.connName       = self.connName
        self.defDataType    = self.defDataType
        self.update         = self.update

        """ DB PROPERTIES """
        self.connUrl = self.setProperties(propKey=eJson.jValues.URL, propVal=connUrl)
        self.connExtraUrl = self.setProperties(propKey=eJson.jValues.URLPARAM, propVal=connExtraUrl)

        self.connObj    = self.setProperties (propKey=eJson.jValues.OBJ, propVal=connObj)
        self.connFilter = self.setProperties (propKey=eJson.jValues.FILTER, propVal=connFilter)
        self.connIsSrc  = self.setProperties (propKey=eJson.jValues.IS_SOURCE, propVal=connIsSrc)
        self.connIsTar  = self.setProperties (propKey=eJson.jValues.IS_TARGET, propVal=connIsTar)
        self.connIsSql  = self.setProperties (propKey=eJson.jValues.IS_SQL, propVal=connIsSql)
        self.sqlFolder  = self.setProperties (propKey=eJson.jValues.FOLDER, propVal=None)
        self.sqlFullFile= self.setProperties(propKey=eJson.jValues.URL_FILE, propVal=None)

        self.defaultSchema  = self.DEFAULTS[eJson.jValues.SCHEMA]
        self.defaulNull     = self.DEFAULTS[eJson.jValues.EMPTY]
        self.defaultSP      = self.DEFAULTS[eJson.jValues.SP]
        self.columnFrame    = self.DEFAULTS[eJson.jValues.COLFRAME]

        self.cursor         = None
        self.connDB         = None
        self.connSql        = None

        self.isExtractSqlIsOnlySTR  = False
        self.parrallelProcessing    = False

        if not self.connUrl:
            self.connUrl = connPropDict

        if self.connIsSql:
            self.connSql    = self.setQueryWithParams(self.connObj)
            self.connObj    = self.connSql

        elif self.connObj and len(self.connObj)>0 and '.sql' not in self.connObj and not self.sqlFullFile:
            self.connSql = "SELECT * FROM %s" %self.connObj

            self.connObj        = self.wrapColName(col=self.connObj, remove=True).split(".")
            self.defaultSchema  = self.connObj[0] if len(self.connObj) > 1 else self.defaultSchema
            self.connObj        = self.connObj[1] if len(self.connObj) > 1 else self.connObj[0]

            if self.connFilter and len(self.connFilter) > 1:
                self.connFilter = re.sub(r'WHERE', '', self.connFilter, flags=re.IGNORECASE)
                self.connSql = '%s WHERE %s' %(self.connSql, self.setQueryWithParams(self.connFilter))

        self.__mapObjectToFile()
        objName = "QUERY " if self.connIsSql else "TABLE: %s" %self.connObj

        self.connect()
        p("CONNECTED, DB TYPE: %s, %s" % (self.conn, objName, ), "ii")

    """  MANADATORY METHOD INHERTIED FROM BASE BATCH INTERFACE"""

    def connect(self):
        try:
            if eConn.MYSQL == self.conn:
                import pymysql
                self.connDB = pymysql.connect(self.connUrl["host"], self.connUrl["user"], self.connUrl["passwd"],
                                              self.connUrl["db"])
                self.cursor = self.connDB.cursor()
            elif eConn.VERTICA == self.conn:
                import vertica_python
                self.connDB = vertica_python.connect(self.connUrl)
                self.cursor = self.connDB.cursor()
            elif eConn.ORACLE == self.conn:
                self.isExtractSqlIsOnlySTR = True
                self.connDB = cx_Oracle.connect(self.connUrl['user'], self.connUrl['pass'], self.connUrl['dsn'])
                if 'nls' in self.connUrl:
                    os.environ["NLS_LANG"] = self.connUrl['nls']
                self.cursor = self.connDB.cursor()
            elif eConn.ACCESS == self.conn:
                import pyodbc as odbc
                self.connDB = odbc.connect(self.connUrl)  # , ansi=True
                self.cursor = self.conn.cursor()
                self.cColoumnAs = False
            elif eConn.LITE == self.conn:
                import sqlite3 as sqlite
                self.connDB = sqlite.connect(self.connUrl)  # , ansi=True
                self.cursor = self.connDB.cursor()
            else:
                try:
                    import ceODBC as odbc
                except ImportError:
                    # p("ceODBC is not installed will try to load pyodbc", "ii")
                    try:
                        import pyodbc as odbc
                    except ImportError:
                        p("pyobbc is not installed", "ii")
                if odbc:
                    self.connDB = odbc.connect(self.connUrl)  # ansi=True
                    self.cursor = self.connDB.cursor()
            return True
        except ImportError:
            p("%s is not installed" % (self.conn), "e")
        except Exception as e:
            err = "Error connecting into DB: %s, ERROR: %s\n " % (self.conn, str(e))
            err += "USING URL: %s\n" %(self.connUrl)
            #err+= traceback.format_exc()
            raise ValueError(err)

    def close(self):
        try:
            if self.cursor:
                self.cursor.close()
            if self.connDB:
                self.connDB.close()
            self.connDB   = None
            self.cursor = None
        except Exception as e:
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            p("baseConnDb->Close: ERROR: file name:"+str(fname)+" line: "+str(exc_tb.tb_lineno)+" massage: "+str(exc_obj), "e")

    def create(self, stt=None, objName=None, addIndex=None):
        tableSchema, tableName = self.setTableAndSchema(tableName=objName, tableSchema=None, wrapTable=False)
        tableFullName = '%s.%s'%(tableSchema, tableName) if tableSchema else tableName

        if not stt or len(stt) == 0:
            p("TABLE %s NOT MAPPED CORRECLTY " %(self.connObj), "e")
            return

        isNew,isExists, newHistoryTable = self.cloneObject(stt, tableSchema, tableName)

        if isNew or  (isExists and self.update<2):
            sql = "CREATE TABLE %s \n (" %(tableFullName)
            for col in stt:
                if eJson.jSttValues.ALIACE in stt[col] and stt[col][eJson.jSttValues.ALIACE] and len (stt[col][eJson.jSttValues.ALIACE])>0:
                    colName = self.wrapColName (col=stt[col][eJson.jSttValues.ALIACE], remove=True)
                else:
                    colName =  self.wrapColName (col=col, remove=True)
                colType =  stt[col][eJson.jSttValues.TYPE]
                sql += '\t%s\t%s,\n' %(colName, colType)

            sql = sql[:-2]+')'
            p("CREATE TABLE: \n" + sql)
            self.exeSQL(sql=sql, commit=True)

        # Check for index
        if addIndex and self.update<2:
            self.addIndexToTable(tableName, addIndex)

        if self.update == 1:
            if newHistoryTable and len (newHistoryTable)>0:
                columns = []
                pre, pos = self.columnFrame[0], self.columnFrame[1]
                oldStructure = self.getStructure(tableName=newHistoryTable, tableSchema=tableSchema, sqlQuery=None)
                newStructure = self.getStructure(tableName=tableName, tableSchema=tableSchema, sqlQuery=None)

                oldStructureL = {x.replace(pre, "").replace(pos, "").lower(): x for x in oldStructure}
                newStructureL = {x.replace(pre, "").replace(pos, "").lower(): x for x in newStructure}

                for col in oldStructureL:
                    if col in newStructureL:
                        columns.append (self.wrapColName(col, remove=False))

                if len (columns)>0:
                    sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.TABLE_COPY_BY_COLUMN, tableName=tableName,tableSchema=tableSchema, srcTableName=newHistoryTable, columns=columns )
                    self.exeSQL(sql=sql)
                    p("ADD ROWS TO %s FROM %s, UPDATED COLUMNS:\n%s" %(tableName, newHistoryTable, ",".join(columns)), "w")

    """ INTERFACE: baseConn, baseConnDB, IMPLEMENTED: db, Method:getStrucutre - return Structure dictionary 
        Strucutre Dictinary: {Column Name: {ColumnType:XXXXX, ColumnAliace: NewColumns } .... } 
        Help methods: __getAccessStructure (db), getQueryStructure (baseConnDB), getDBStructure (baseConnDB)"""
    def getStructure(self, tableName=None, tableSchema=None, sqlQuery=None):
        sqlQuery = sqlQuery if sqlQuery else self.connSql
        tableSchema, tableName = self.setTableAndSchema (tableName=tableName, tableSchema=tableSchema)
        retStructure = None
        pre,pos = self.columnFrame[0],self.columnFrame[1]


        # If there is query and there is internal maaping in query - will add this mapping to mappingColum dictionary
        if self.connIsSql:
            retStructure = self.getQueryStructure(sqlQuery=sqlQuery)

        elif tableName and len(tableName) > 0:
            retStructure = self.getDBStructure(tableSchema=tableSchema, tableName=tableName)

        finallStructure = OrderedDict()
        for col in  retStructure:
            #col = col.replace (pre,"").replace(pos,"")
            finallStructure[col] = retStructure[col]

            if eJson.jSttValues.ALIACE in finallStructure[col] and finallStructure[col][eJson.jSttValues.ALIACE] is not None:
                finallStructure[col][eJson.jSttValues.ALIACE] = finallStructure[col][eJson.jSttValues.ALIACE].replace(pre,"").replace(pos,"")

        return finallStructure

    def addIndexToTable (self, tableName, addIndex, tableSchema=None):
        if addIndex and len(addIndex)>0:
            newIndexList    = []
            existIndexDict  = {}
            isClusterExists = False
            toCreate        = True

            ## update existing index
            addIndex = [addIndex] if isinstance(addIndex, (dict,OrderedDict)) else addIndex
            for ind in addIndex:
                if eJson.jValues.INDEX_COLUMS in ind and ind[eJson.jValues.INDEX_COLUMS]:
                    columns  = ind[eJson.jValues.INDEX_COLUMS] if isinstance(ind[eJson.jValues.INDEX_COLUMS], list) else [ind[eJson.jValues.INDEX_COLUMS]]
                    columns  = [x.lower() for x in columns]
                    isCluster= ind[eJson.jValues.INDEX_CLUSTER] if eJson.jValues.INDEX_CLUSTER in ind and ind[eJson.jValues.INDEX_CLUSTER] is not None else False
                    isUnique = ind[eJson.jValues.INDEX_UNIQUE] if eJson.jValues.INDEX_UNIQUE in ind and ind[eJson.jValues.INDEX_UNIQUE] is not None else False
                    newIndexList.append ({eJson.jValues.INDEX_COLUMS:columns, eJson.jValues.INDEX_CLUSTER:isCluster, eJson.jValues.INDEX_UNIQUE:isUnique})
                else:
                    p("NOT VALID INDEX %s , MUST HAVE COLUNs, IGNORING" %(str(ind)), "i")


            ## update newIndexDict from db.
            tableSchema, tableName = self.setTableAndSchema(tableName=tableName, tableSchema=tableSchema)
            tableName = '%s.%s' %(tableSchema,tableName) if tableSchema else tableName
            # check if there is cluster index
            sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.INDEX_EXISTS, tableName=tableName)
            self.cursor.execute(sql)
            rows = self.cursor.fetchall()
            if rows and len (rows)>0:
                for row in rows:
                    indexName = row[0]
                    if indexName not in existIndexDict:
                        existIndexDict[indexName] = {eJson.jValues.INDEX_COLUMS:[], eJson.jValues.INDEX_CLUSTER:False,eJson.jValues.INDEX_UNIQUE:False}

                    existIndexDict[indexName][eJson.jValues.INDEX_CLUSTER] = True if row[2] or str(row[2]) == '1' else False
                    if existIndexDict[indexName][eJson.jValues.INDEX_CLUSTER] == True:
                        isClusterExists = True
                    existIndexDict[indexName][eJson.jValues.INDEX_UNIQUE] = True if row[3] or str(row[3])  == '1' else False
                    existIndexDict[indexName][eJson.jValues.INDEX_COLUMS].append (str(row[1]).lower())

            ### Compare - Remove existing idential Indexes
            for eInd in existIndexDict:
                if existIndexDict[eInd] in newIndexList:
                    newIndexList.remove(existIndexDict[eInd])

            ### Compare index with existing indexes
            for ind in newIndexList:
                if ind[eJson.jValues.INDEX_CLUSTER] == True and isClusterExists:
                    p("CLUSTERED INDEX ALREADY EXISTS, IGNORE NEW INDEX:%s" %(str(ind)), "w")
                    toCreate = False
                else:
                    newColumn = ind[eJson.jValues.INDEX_COLUMS]

                    for eInd in existIndexDict:
                        eColumn = eInd[eJson.jValues.INDEX_COLUMS]

                        if set(eColumn) == set(newColumn):
                            p("INDEX ON COLUMN %s EXISTS, OLD IS_CLUSTER:%s, OLD IS_UNIQUE:%s, NEW IS_CLUSTER: %s, NEW IS_UNIQUE: %s, IGNORING..." %(eColumn, eInd[eJson.jValues.INDEX_CLUSTER], eInd[eJson.jValues.INDEX_UNIQUE], ind[eJson.jValues.INDEX_CLUSTER], ind[eJson.jValues.INDEX_UNIQUE] ) , "i")
                            toCreate = False
                    if toCreate:
                        columnsCreate   = ind[eJson.jValues.INDEX_COLUMS]
                        isUnique        = ind[eJson.jValues.INDEX_UNIQUE]
                        isCluster       = ind[eJson.jValues.INDEX_CLUSTER]
                        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.INDEX, tableName=tableName, columns=columnsCreate, isCluster=isCluster, isUnique=isUnique)
                        print ("TAL 8899", sql)
                        self.exeSQL(sql=sql)
                        p("TYPE:%s, ADD INDEX: COLUNS:%s, CLUSTER:%s, UNIQUE:%s" % (self.conn, str(columnsCreate),str(isCluster),str(isUnique)), "ii")

    def isExists(self, tableName, tableSchema):
        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.ISEXISTS, tableName=tableName, tableSchema=tableSchema)
        self.cursor.execute(sql)
        row = self.cursor.fetchone()
        if row and row[0]:
            p("baseConnDb->isExists SCHEMA:%s, TABLE:%s EXISTS" %(tableSchema, tableName), "ii")
            return True
        p("baseConnDb->isExists: SCHEMA:%s, TABLE:%s NOT EXISTS" % (tableSchema, tableName), "ii")
        return False

    def preLoading(self, tableName=None, tableSchema=None, sqlFilter=None):
        sqlFilter = sqlFilter if sqlFilter else self.connFilter
        if sqlFilter and len(sqlFilter) > 0:
            self.delete(sqlFilter=sqlFilter, tableName=tableName, tableSchema=tableSchema)
        else:
            self.truncate(tableName=tableName, tableSchema=tableSchema)

    def extract(self, tar, tarToSrc, batchRows=None, addAsTaret=True):
        batchRows = batchRows if batchRows else self.batchSize
        fnOnRowsDic     = {}
        execOnRowsDic   = {}
        pre,pos         = self.columnFrame[0],self.columnFrame[1]
        sourceColumnStr = []
        targetColumnStr = []
        sourceSql       = self.connSql

        ## There is Source And Target column mapping
        if tarToSrc and len (tarToSrc)>0:
            existingColumnsDic = OrderedDict()
            existingColumns = qp.extract_tableAndColumns(sql=sourceSql)
            preSql = existingColumns[qp.QUERY_PRE]
            postsql = existingColumns[qp.QUERY_POST]
            if self.connIsSql:
                allColumns  = existingColumns[qp.QUERY_COLUMNS_KEY]
                pre, pos = self.columnFrame[0], self.columnFrame[1]

                for col in allColumns:
                    existingColumnsDic[col[1].replace(pre,"").replace(pos,"").lower()] = ".".join(col[0])
            else:
                allColumns = self.getStructure()
                for col in allColumns:
                    existingColumnsDic[col.replace(pre,"").replace(pos,"").lower()] = col

            for i,col in  enumerate (tarToSrc):
                colTarName = '%s%s%s' % (pre, col.replace(pre, "").replace(pos, ""), pos)
                if col.replace(pre,"").replace(pos,"").lower() in existingColumnsDic:
                    colSrcName = existingColumnsDic[col.replace(pre,"").replace(pos,"").lower()]
                    colSrcName = '%s As %s' % (colSrcName, colTarName) if addAsTaret else colSrcName
                else:
                    colSrcName =  "'' As %s" %(colTarName) if addAsTaret else ''

                sourceColumnStr.append (colSrcName)
                targetColumnStr.append (col)

                ### ADD FUNCTION
                if eJson.jSttValues.FUNCTION in tarToSrc[col] and tarToSrc[col][eJson.jSttValues.FUNCTION]:
                    fnc = eval(tarToSrc[col][eJson.jSttValues.FUNCTION])
                    fnOnRowsDic[i] = fnc if isinstance(fnc, (list, tuple)) else [fnc]

                ### ADD EXECUTION FUNCTIONS
                elif eJson.jSttValues.EXECFUNC in tarToSrc[col] and len(tarToSrc[col][eJson.jSttValues.EXECFUNC])>0:
                    newExcecFunction = tarToSrc[col][eJson.jSttValues.EXECFUNC]
                    regex   = r"(\{.*?\})"
                    matches = re.finditer(regex, tarToSrc[col][eJson.jSttValues.EXECFUNC], re.MULTILINE | re.DOTALL)
                    for matchNum, match in enumerate(matches):
                        for groupNum in range(0, len(match.groups())):
                            colName = match.group(1)
                            if colName and len(colName)>0:
                                colToReplace = match.group(1).replace("{","").replace("}","")
                                colToReplace = self.__isColumnExists (colName=colToReplace, tarToSrc=tarToSrc)
                                if colToReplace:
                                    newExcecFunction = newExcecFunction.replace(colName,"{"+str(colToReplace)+"}")
                    execOnRowsDic[i] = newExcecFunction

            columnStr = ",".join(sourceColumnStr)
            sourceSql = '%s %s %s' %(preSql,columnStr,postsql)

        """ EXECUTING SOURCE QUERY """
        sourceSql = str(sourceSql) if self.isExtractSqlIsOnlySTR else sourceSql

        self.exeSQL(sql=sourceSql , commit=False)
        p("EXTRACTING SQL:\n %s" %sourceSql,"ii")

        if len(targetColumnStr) == 0:
            targetColumnStr = [col[0] for col in self.cursor.description]

        rows = None
        try:
            if batchRows and batchRows>0:
                while True:

                    rows = self.cursor.fetchmany( batchRows )
                    if not rows or len(rows) < 1:
                        break
                    rows = self.dataTransform(data=rows, functionDict=fnOnRowsDic, execDict=execOnRowsDic)
                    tar.load (rows=rows, targetColumn = targetColumnStr)
            else:
                rows = self.cursor.fetchall()
                rows = self.dataTransform(data=rows, functionDict=fnOnRowsDic)
                tar.load(rows, targetColumn = targetColumnStr)
        except Exception as e:
            p("TYPE:%s, OBJECT:%s ERROR FATCHING DATA" % (self.conn, str(self.connObj)), "e")
            p(str(e), "e")

    def load(self, rows, targetColumn):
        totalRows = len(rows) if rows else 0
        if totalRows == 0:
            p("THERE ARE NO ROWS")
            return
        tblFullName = "%s.%s" %(self.defaultSchema, self.connObj) if self.defaultSchema else self.connObj
        execQuery = "INSERT INTO %s" % (tblFullName)
        pre, pos = self.columnFrame[0], self.columnFrame[1]
        colList = []
        colInsert = []
        ## Compare existint target strucutre
        tarStrucutre = self.getStructure(tableName=self.connObj, tableSchema=self.defaultSchema, sqlQuery=None)
        tarStrucutreL= {x.replace(pre, "").replace(pos, "").lower(): x for x in tarStrucutre}

        removeCol = {}
        for i, col in enumerate (targetColumn):
            if col.replace(pre, "").replace(pos, "").lower() not in tarStrucutreL:
                removeCol[i] = col
        if len (removeCol)>0:
            for num in removeCol:
                p("COLUMN NUMBER %s, NAME: %s NOT EXISTS IN TARGET TABLE, IGNORE COLUMN" %(num,removeCol[num]), "w")
                targetColumn.remove(removeCol[num])

                for i, r in enumerate (rows):
                    rows[i] = list(r)
                    del rows[i][num]

        for col in targetColumn:
            colName = '%s%s%s' % (pre, col, pos)
            colList.append(colName)
            colInsert.append('?')
        execQuery += "(%s) " % (",".join(colList))
        execQuery += "VALUES (%s)" % (",".join(colInsert))

        try:
            self.cursor.executemany(execQuery, rows)
            self.connDB.commit()
            p('LOAD %s into target: %s >>>>>> ' % (str(totalRows), self.connObj), "ii")


        except Exception as e:
            p(u"TYPE:%s, OBJCT:%s ERROR in cursor.executemany !!!!" % (self.conn, self.connObj), "e")
            p(u"ERROR QUERY:%s " % execQuery, "e")
            sampleRes = ['Null' if not r else "'%s'" % r for r in rows[0]]
            p(u"SAMPLE:%s " % u", ".join(sampleRes), "e")
            p(e, "e")
            if config.LOOP_ON_ERROR:
                iCnt = 0
                tCnt = len(rows)
                errDict = {}
                totalErrorToLooap = int(tCnt * 0.1)
                totalErrorsFound = 0
                p("ROW BY ROW ERROR-> LOADING %s OUT OF %s ROWS " % (str(totalErrorToLooap), str(tCnt)),
                  "e")
                for r in rows:
                    try:
                        iCnt += 1
                        r = [r]
                        self.cursor.executemany(execQuery, r)
                        self.connDB.commit()
                    except Exception as e:
                        totalErrorsFound += 1
                        if totalErrorsFound > totalErrorToLooap:
                            break
                        errMsg = str(e).lower()

                        if errMsg not in errDict:
                            errDict[errMsg] = 0
                            ret = ""
                            for col in r[0]:
                                if col is None:
                                    ret += "Null, "
                                else:
                                    ret += "'%s'," % (col)
                            p("ROW BY ROW ERROR-> %s" %execQuery, "e")
                            p(ret, "e")
                            p(e, "e")
                        else:
                            errDict[errMsg] += 1

                p("ROW BY ROW ERROR-> LOADED %s OUT OF %s ROWS" % (str(totalErrorToLooap), str(tCnt)), "e")
                for err in errDict:
                    p("TOTAL ERRORS: %s, MSG: %s: " % (str(err), str(errDict[err])), "e")

    def execMethod(self, method=None):
        method = method if method else self.connObj

        if method and len(method)>0:
            p("CONN:%s, EXEC METHOD:\n%s" %(self.conn, method), "i")
            methodTup = [(1,method,{})]
            execQuery(sqlWithParamList=method, connObj=self)

    """ PUBLIC METHOD FOR DB MANIPULATION  """
    def exeSQL(self, sql, commit=True):
        s = ''
        if not (isinstance(sql, (list, tuple))):
            sql = [sql]
        try:
            for s in sql:
                self.cursor.execute(s)  # if 'ceodbc' in odbc.__name__.lower() else self.conn.execute(s)
            if commit:
                self.connDB.commit()  # if 'ceodbc' in odbc.__name__.lower() else self.cursor.commit()
            return True
        except Exception as e:
            p(e, "e")
            p(u"ERROR SQL:\n%s " % (uniocdeStr(s)), "e")
            return False

    """ baseConnDB Method - Truncate table """
    def truncate(self, tableName=None, tableSchema=None):
        tableSchema, tableName  = self.setTableAndSchema (tableName=tableName, tableSchema=tableSchema, wrapTable=True)
        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.TRUNCATE, tableName=tableName, tableSchema=tableSchema)
        self.exeSQL(sql=sql)
        p("TYPE:%s, TRUNCATE TABLE:%s" % (self.conn, self.connObj),"ii")

    """ baseConnDB Method - Delete rows from table by using filter paramters"""
    def delete (self, sqlFilter, tableName=None, tableSchema=None):
        tableSchema, tableName = self.setTableAndSchema(tableName=tableName, tableSchema=tableSchema, wrapTable=True)
        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.DELETE, tableName=tableName, tableSchema=tableSchema, sqlFilter=sqlFilter)
        self.exeSQL(sql=sql)
        p("TYPE:%s, DELETE FROM TABLE:%s, WHERE:%s" % (self.conn, self.connObj, self.connFilter), "ii")

    """ Return TABLE STRUCTURE : {ColumnName:{Type:ColumnType, ALIACE: ColumnName} .... } """
    def getDBStructure(self, tableSchema, tableName):
        ret = OrderedDict()

        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.STRUCTURE, tableName=tableName, tableSchema=tableSchema)

        self.exeSQL(sql, commit=False)
        rows = self.cursor.fetchall()

        for col in rows:
            colName = uniocdeStr(col[0])
            colType = uniocdeStr(col[1])
            val = {eJson.jSttValues.TYPE: colType, eJson.jSttValues.ALIACE: None}
            ret[colName] = val
        return ret

    """ Return complex or simple QUERY STRUCURE:  {ColumnName:{Type:ColumnType, ALIACE: ColumnName} .... } """
    def getQueryStructure(self, sqlQuery=None):

        sqlQuery = sqlQuery if sqlQuery else self.connSql
        ret = OrderedDict()
        pre = self.columnFrame[0]
        pos = self.columnFrame[1]

        notFoundColumns = {}
        allFoundColumns = []
        foundColumns = []

        if not sqlQuery or len(sqlQuery) < 1:
            return ret

        ### Return dictionary : {Table Name:[{SOURCE:ColumnName, ALIASE: column aliase}, ....]}
        ### And empty table -> all column that not related to any table '':[{SOURCE:columnName, ALIASE: .... } ...]
        queryTableAndColunDic = qp.extract_tableAndColumns(sql=sqlQuery)

        if qp.QUERY_COLUMNS_KEY in queryTableAndColunDic:
            allFoundColumns = queryTableAndColunDic[qp.QUERY_COLUMNS_KEY]
            del queryTableAndColunDic[qp.QUERY_COLUMNS_KEY]


        if qp.QUERY_NO_TABLE in queryTableAndColunDic:
            for colD in queryTableAndColunDic[qp.QUERY_NO_TABLE][qp.TABLE_COLUMN]:
                colTupleName = colD[0]
                colTargetName= colD[1]
                colName = colTupleName[-1]
                colFullName =".".join(colTupleName)
                notFoundColumns[colTargetName] = (colName, colFullName, )


            del queryTableAndColunDic[qp.QUERY_NO_TABLE]

        if qp.QUERY_POST in queryTableAndColunDic:
            del queryTableAndColunDic[qp.QUERY_POST]

        if qp.QUERY_PRE in queryTableAndColunDic:
            del queryTableAndColunDic[qp.QUERY_PRE]

        # update allTableStrucure dictionary : {tblName:{col name : ([original col name] , [tbl name] , [col structure])}}
        for tbl in queryTableAndColunDic:
            colDict = OrderedDict()
            tableName = tbl
            tableSchema = queryTableAndColunDic[tbl][qp.TABLE_SCHEMA]
            tableColumns= queryTableAndColunDic[tbl][qp.TABLE_COLUMN]


            for colD in tableColumns:
                colTupleName = colD[0]
                colTargetName= colD[1]
                colName = colTupleName[-1]
                colFullName =".".join(colTupleName)
                colDict[colTargetName] = (colName, colFullName)

            tableStrucure = self.getDBStructure(tableSchema=tableSchema, tableName=tableName)
            tableColL = {x.replace(pre, "").replace(pos, "").lower(): x for x in tableStrucure}

            for col in colDict:
                colTarName = col
                colSName = colDict[col][0]
                colLName = colDict[col][1]
                isFound = False
                if '*' in colSName:
                    isFound = True
                    ret.update(tableStrucure)
                else:
                    for colTbl in tableColL:
                        if colTbl.lower() in colSName.lower():
                            ret[colTarName] = {eJson.jSttValues.SOURCE:tableColL[colTbl], eJson.jSttValues.TYPE:tableStrucure[tableColL[colTbl]][eJson.jSttValues.TYPE]}
                            isFound = True
                            break

                if not isFound:
                    p("COLUMN %s NOT FOUND IN TABLE %s USING DEFAULT %s" % (col, tbl, self.defDataType), "ii")
                    ret[colTarName] = {eJson.jSttValues.SOURCE: colLName,eJson.jSttValues.TYPE: self.defDataType}

            ## Search for Column that ther is no table mapping
            for col in notFoundColumns:
                colTarName = col
                colSName = notFoundColumns[col][0]
                colLName = notFoundColumns[col][1]
                for colTbl in tableColL:
                    if colTbl.lower() in colSName.lower():
                        ret[colTarName] = {eJson.jSttValues.SOURCE: colLName,eJson.jSttValues.TYPE: tableStrucure[tableColL[colTbl]][eJson.jSttValues.TYPE]}
                        foundColumns.append (colLName)
                        break

        for col in notFoundColumns:
            colTarName = col
            colSName = notFoundColumns[col][0]
            colLName = notFoundColumns[col][1]
            if colLName not in foundColumns:
                p("COLUMN %s NOT FOUND IN ANY TABLE, USING DEFAULT %s" % (colLName,  self.defDataType), "ii")
                ret[colTarName] = {eJson.jSttValues.SOURCE: colLName,eJson.jSttValues.TYPE: self.defDataType}

        retOrder = OrderedDict()
        for col in allFoundColumns:
            if col[1] in ret:
                retOrder[col[1]] = ret[col[1]]
            else:
                p("!!!!ERROR: COLUMN %s NOT FOUND " %col[1],"w")
        for col in ret:
            if col not in retOrder:
                p("!!!!ERROR: COLUMN %s NOT FOUND " % col, "w")

        return ret

    """ INTERNAL USED for create method
        Create new Table is table is exsist
        If config.TRACK_HISTORY will save old table as tablename_currentDate   """
    def cloneObject(self, newStructure, tableSchema, tableName):
        tableSchema, tableName = self.setTableAndSchema(tableName=tableName, tableSchema=tableSchema, wrapTable=False)
        schemaEqual = True
        newStructureL = OrderedDict()
        newHistoryTable = None
        pre,pos = self.columnFrame[0], self.columnFrame[1]

        for col in newStructure:
            colAlias = newStructure[col][eJson.jSttValues.ALIACE] if eJson.jSttValues.ALIACE in newStructure[col] else None
            colType  = newStructure[col][eJson.jSttValues.TYPE] if eJson.jSttValues.TYPE in newStructure[col] else self.defDataType
            if colAlias:
                newStructureL[colAlias.replace(pre,"").replace(pos,"").lower()] = (colAlias,colType)
            else:
                newStructureL[col.replace(pre,"").replace(pos,"").lower()] = (col, colType)

        existStructure = self.getStructure (tableName=tableName, tableSchema=tableSchema,sqlQuery=None)

        if not existStructure or len(existStructure) == 0:
            p("TABLE %s NOT EXISTS " %(tableName), "ii")
            return True,False, newHistoryTable

        updateDesc = 'UPDATE' if self.update == 1 else 'DROP CREATE' if self.update < 1 else 'WARNIN, NO UPDATE'

        existStructureL = {x.replace(pre, "").replace(pos, "").lower(): x for x in existStructure}
        for col in existStructureL:
            if col in newStructureL:
                if existStructure[existStructureL[col]][eJson.jSttValues.TYPE].lower() != newStructureL[col][1].lower():
                    schemaEqual = False
                    if self.update == 1:
                        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.TABLE_CHANGE_COLUMN,
                                                   tableName=tableName, tableSchema=tableSchema, columnName=col,
                                                   columnType=newStructureL[col][1])
                        self.exeSQL(sql=sql)

                    p("%s: CONN:%s, TABLE: %s, COLUMN %s, TYPE CHANGED, OLD: %s, NEW: %s" % (updateDesc, self.conn, tableName, col, existStructure[existStructureL[col]][eJson.jSttValues.TYPE],newStructureL[col][1]), "w")

            else:
                schemaEqual = False
                p("CONN:%s, TABLE: %s, REMOVING COLUMN: %s " % (self.conn, tableName, col), "w")

        for col in newStructureL:
            if col not in existStructureL:
                schemaEqual = False
                p("CONN:%s, TABLE: %s, ADD COLUMN: %s " % (self.conn, tableName, newStructureL[col][0]), "w")

        if schemaEqual:
            p("TABLE %s DID NOT CHANGED  >>>>>" % (tableName), "ii")
            return False,False, newHistoryTable
        else:
            if self.update>1:
                p("TABLE STRUCTURE CHANGED, UPDATE IS NOT ALLOWED, NO CHANGE","w")
                return False,True, newHistoryTable
            else:
                if config.TRACK_HISTORY:
                    p("TABLE HISTORY IS ON ...", "ii")
                    newHistoryTable = "%s_%s" % (tableName, str(time.strftime('%y%m%d')))
                    if (self.isExists(tableSchema=tableSchema, tableName=tableName)):
                        num = 0
                        while (self.isExists(tableSchema=tableSchema, tableName=newHistoryTable)):
                            num += 1
                            newHistoryTable = "%s_%s_%s" % (tableName, str(time.strftime('%y%m%d')), str(num))
                    if newHistoryTable:
                        p("TABLE HISTORY IS ON AND CHANGED, TABLE %s EXISTS ... RENAMED TO %s" % (str(tableName), str(newHistoryTable)), "w")
                        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.RENAME, tableSchema=tableSchema,tableName=tableName, tableNewName=newHistoryTable)

                        # sql = eval (self.objType+"_renameTable ("+self.objName+","+oldName+")")
                        p("RENAME TABLE SQL:%s" % (str(sql)), "w")
                        self.exeSQL(sql=sql, commit=True)
                else:
                    if existStructure and len(existStructure)>0:
                        p("TABLE HISTORY IS OFF AND TABLE EXISTS, DROP -> CREATE TABLE %s IN NEW STRUCTURE... "%(str(tableName)), "w")
                        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.DROP,  tableName=tableName, tableSchema=tableSchema)
                        self.exeSQL(sql=sql, commit=True)
        return False,True, newHistoryTable

    """ Return tableSchema, tableName From table name and schema 
        WrapTable=True will return with DB wrapping for example Sql server colum yoyo will be [yoyo] """
    def setTableAndSchema (self, tableName, tableSchema=None, wrapTable=False):
        tableSchema = tableSchema if tableSchema else self.defaultSchema
        tableName   = tableName if tableName else self.connObj

        tableName = tableName.split (".")
        if len(tableName) == 1:
            tableName   = tableName[0]
        else:
            tableSchema = tableName[0]
            tableName   = tableName[1]

        tableSchema = self.wrapColName (tableSchema, remove=not wrapTable)
        tableName   = self.wrapColName (tableName, remove=not wrapTable)
        return tableSchema, tableName

    """ Wrap Column with DB brackets. Sample SqlServer column: productID --> [productId] """
    def wrapColName (self, col, remove=False):
        if self.columnFrame and col and len (col)>0:
            srcPre, srcPost = self.columnFrame
            coList = col.split(".")
            ret = ""
            for col in coList:
                col = col.replace(srcPre,"").replace(srcPost,"")
                if not remove:
                    col= "%s%s%s" %(srcPre,col,srcPost)
                ret+=col+"."
            return ret[:-1]
        return col

    """ Cnvert all paramters in config.QUERY_PARAMS into variable and add it into SQL QUERY """
    def setQueryWithParams(self, query, queryParams=None):
        if queryParams and config.QUERY_PARAMS:
            queryParams.update(config.QUERY_PARAMS)
        else:
            queryParams = config.QUERY_PARAMS
        qRet = u""
        if query and len(query) > 0:
            if isinstance(query, (list, tuple)):
                for q in query:
                    for param in queryParams:
                        q = self.__replaceStr(sString=q, findStr=param, repStr=queryParams[param], ignoreCase=True,addQuotes="'")
                    qRet += q + u" "
            else:
                for param in queryParams:
                    if param in query:
                        query = self.__replaceStr(sString=query, findStr=param, repStr=queryParams[param],ignoreCase=True, addQuotes="'")
                qRet += query
            if len(queryParams)>0:
                p("REPLACE PARAMS: %s" % (str(queryParams)), "ii")
        else:
            qRet = query
        return qRet

    """ INTERNAL METHOD  """

    """ Internal Method for replace paramters in setQueryWithParams Metod """
    def __replaceStr(self, sString, findStr, repStr, ignoreCase=True, addQuotes=None):
        if addQuotes and isinstance(repStr, str):
            findStrWithQuotes = "%s%s%s" % (addQuotes, findStr, addQuotes)
            if findStrWithQuotes not in sString:
                repStr = "%s%s%s" % (addQuotes, repStr, addQuotes)

        if ignoreCase:
            pattern = re.compile(re.escape(findStr), re.IGNORECASE)
            res = pattern.sub(repStr, sString)
        else:
            res = sString.replace(findStr, repStr)
        return res

    def __isColumnExists (self, colName, tarToSrc):
        for ind, col in enumerate (tarToSrc):
            if col.lower() == colName.lower():
                return ind
            elif eJson.jSttValues.SOURCE in tarToSrc[col] and tarToSrc[col][eJson.jSttValues.SOURCE].lower() == col.lower():
                return ind

        p("COLUMN %s NOT FOUND IN MAPPING" %(colName))
        return None

    def __mapObjectToFile (self):
        fullFilePath = None
        if self.sqlFullFile and os.path.isfile(self.sqlFullFile):
            fullFilePath = self.sqlFullFile
        elif self.sqlFolder and self.sqlFullFile and  os.path.isfile( os.path.join (self.sqlFolder,self.sqlFullFile)):
            fullFilePath = os.path.join (self.sqlFolder,self.sqlFullFile)
        elif self.sqlFolder and os.path.isfile(os.path.join(self.sqlFolder, self.connObj)):
            fullFilePath = self.sqlFullFile(os.path.join(self.sqlFolder, self.connObj))

        if fullFilePath:
            foundQuery= False
            allParams = []
            if '.sql' in self.connObj:
                self.connObj = fullFilePath
            else:
                with io.open(fullFilePath, 'r', encoding='utf8') as inp:
                    sqlScript = inp.readlines()
                    allQueries = qp.querySqriptParserIntoList(sqlScript, getPython=True, removeContent=True, dicProp=None)

                    for q in allQueries:
                        allParams.append(q[1])
                        if q[0] and q[0].lower() == str(self.connObj).lower():
                            p("USING %s AS SQL QUERY " %(q[0]),"ii")
                            self.connObj = q[1]
                            self.connIsSql = True
                            foundQuery = True
                            break
                    else:
                        p("USING %s DIRECTLY, NOT FOUND IN %s" %(self.connObj,fullFilePath), "ii")

    def minValues (self, colToFilter=None, resolution=None, periods=None, startDate=None):
        raise NotImplementedError("minValues need to be implemented")

    def merge (self, mergeTable, mergeKeys=None, sourceTable=None):
        srcSchema, srcName = self.setTableAndSchema(tableName=sourceTable, tableSchema=None, wrapTable=True)
        mrgSchema, mrgName = self.setTableAndSchema(tableName=mergeTable, tableSchema=None, wrapTable=True)

        srcStructure    = self.getDBStructure(tableSchema=srcSchema, tableName=srcName)
        mrgStructure    = self.getDBStructure(tableSchema=mrgSchema, tableName=mrgName)

        mrgStructureL   = {x.lower():x for x in mrgStructure}
        srcStructureL   = {x.lower():x for x in srcStructure}
        mergeKeysL      = [self.wrapColName(col=x.lower(), remove=False) for x in mergeKeys]

        ### MERGE IDENTICAL COLUMN ONLY
        updateColumns       = []
        keyColumns          = []
        allColumns          = []
        notExistsInMergeCol = []
        notExistsInSourceCol= []

        for col in srcStructure:
            if col.lower() in mrgStructureL:
                col = self.wrapColName(col=col, remove=False)
                allColumns.append (col)
                if col.lower() in mergeKeysL:
                    keyColumns.append(col)
                else:
                    updateColumns.append(col)
            else:
                notExistsInMergeCol.append (col)

        for col in mrgStructure:
            if col.lower() not in srcStructureL:
                notExistsInSourceCol.append (col)

        if len (notExistsInMergeCol)>0:
            p ("SOURCE COLUMNS %s NOT EXISTS IN MERGE TABLE %s" %(str(notExistsInMergeCol), sourceTable),"ii")

        if len (notExistsInSourceCol)>0:
            p ("MERGE COLUMNS %s NOT EXISTS IN SOURCE TABLE %s" %(str(notExistsInSourceCol), mergeTable),"ii")

        if len (keyColumns) == 0:
            keyColumns = updateColumns

        dstTable = '%s.%s' %(mrgSchema,mrgName) if mrgSchema else mrgName
        srcTable = '%s.%s' %(srcSchema,srcName)

        sql = setSqlQuery().getSql(conn=self.conn, sqlType=eSql.MERGE, dstTable=dstTable, srcTable=srcTable, mergeKeys=keyColumns, colList=updateColumns, colFullList=allColumns)
        self.exeSQL(sql=sql)
        p("TYPE:%s, MERGE %s WITH %s, \n\t\tMERGE KEYS:%s" %(self.conn, srcTable, dstTable, str(keyColumns)), "ii")

    def cntRows (self):
        raise NotImplementedError("count rows need to be implemented")


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import  (absolute_import, division, print_function)
__metaclass__ = type

import logging

class config:
    CONN_URL    =  {    'sql'    :"DRIVER={SQL Server};SERVER=server,1433;DATABASE=database;UID=uid;PWD=pass;",
                        'oracle' :"DRIVER={SQL Server};SERVER=server,1433;DATABASE=database;UID=uid;PWD=pass;",
                        'mysql'  :"host=host, user=user, passwd=pass, db=db",
                        'vertica':"DRIVER=HPVertica;SERVER=server;DATABASE=database;PORT=5433;UID=user;PWD=pass",
                        'file'   :{'delimiter':',','header':True, 'folder':""}
                   }

    QUERY_PARAMS    = {}
    SQL_FOLDER_DIR  = None
    PARSER_SQL_MAIN_KEY = "dingDong"

    DECODE          = "windows-1255"
    TRACK_HISTORY   = True
    LOOP_ON_ERROR   = True
    NUM_OF_PROCESSES= 4

    #LOGGING Properties
    LOGS_DEBUG = logging.DEBUG
    LOGS_DIR   = None
    LOGS_INFO_NAME = 'log'
    LOGS_ERR_NAME  = 'log'
    LOGS_TMP_NAME  = 'lastLog'
    LOGS_HISTORY_DAYS=5

    #SMTP and massaging Configuration
    SMTP_SERVER             = ""
    SMTP_SERVER_USER        = ""
    SMTP_SERVER_PASS        = ""
    SMTP_SENDER             = ""
    SMTP_RECEIVERS          = ['info@biSkilled.com']


# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import  (absolute_import, division, print_function)
__metaclass__ = type

import os
import re
import io
import sys
import multiprocessing
from collections import OrderedDict
from past.builtins import basestring

from dingDong.config        import config
from dingDong.misc.logger   import p, LOGGER_OBJECT
from dingDong.misc.misc     import uniocdeStr

if 2 == sys.version_info[0]:
    reload(sys)
    sys.setdefaultencoding("utf8")

# sqlWithParamList --> list of tuple (file name, paramas)
def execQuery (sqlWithParamList, connObj ):
    if sqlWithParamList is None or len(sqlWithParamList)==0:
        p("NOT RECIAVE ANY SQL STATEMENT")
        return

    if isinstance(sqlWithParamList, basestring):
        sqlWithParamList = [(sqlWithParamList,{})]

    allFiles    = OrderedDict()
    sqlFiles    = []
    locParams   = {}

    for script in sqlWithParamList:
        parallelProcess = -1
        if len (script) == 3:
            parallelProcess = script[0]
            locName         = script[1]
            locParams       = script[2]
        elif len (script) == 2:
            locName         = script[0]
            locParams       = script[1]
        elif len (script) == 1:
            locName         = script[0]
        else:
            p("NOT CONFIGURE PROPERLY, MUST HAVE 1, 2 or 3 in a tuple  %s " % (str(script)) ,"e")
            break

        # sql file is list of all files to execute

        if os.path.isdir(locName):
            sqlFiles = [os.path.join(locName, pos_sql)  for pos_sql in os.listdir(locName) if pos_sql.endswith('.sql')]
        elif os.path.isfile(locName):
            sqlFiles.append(locName)
        else:
            sqlFiles.append(locName)

        # Adding all script into ordered dictionary

        if parallelProcess not in allFiles:
            allFiles[parallelProcess] = []

        allFiles[parallelProcess].append ( (sqlFiles, locParams) )
        sqlFiles = list([])

    for priority  in OrderedDict (sorted (allFiles.items())):

        p ('START EXECUTING PRIORITY %s >>>>>>> ' %str(priority), "ii")
        __execParallel (priority, allFiles[priority], connObj)

def __execParallel (priority, ListOftupleFiles, connObj):
    multiProcessParam = []
    multiProcessFiles = ''


    for tupleFiles in ListOftupleFiles:
        sqlFiles    = tupleFiles[0]
        locParams   = tupleFiles[1]


        for sqlScript in sqlFiles:
            multiProcessParam.append( (sqlScript, locParams, connObj, config.LOGS_DEBUG) )
            multiProcessFiles += "'" + sqlScript + "' ; "

    # single process
    if priority<0 or len(multiProcessParam)<2:
        p("SINGLE PROCESS: %s" % (str(multiProcessFiles)), "ii")
        for query in multiProcessParam:
            __execSql(query)

    # multiprocess execution
    else:
        if len(multiProcessParam) > 1:
            p ("MULTI PROCESS: %s" %(str(multiProcessFiles)), "ii")
            # Strat runing all processes
            proc = multiprocessing.Pool(config.NUM_OF_PROCESSES).map( __execSql ,multiProcessParam )

    p("FINISH EXECUTING PRIORITY %s, LOADED FILES: %s >>>> " %(str(priority), str (multiProcessFiles)), "i")

def __execSql ( params ):
    (sqlScript, locParams, connObj, logLevel) = params

    LOGGER_OBJECT.setLogLevel(logLevel=logLevel)

    def __execEachLine (connObj, sqlTxt):
        sqlQuery = __split_sql_expressions(sqlTxt)
        isParam = True if len(locParams) > 0 else False
        for sql in sqlQuery:
            if isParam:
                sql =  connObj.setQueryWithParams (query=sql, queryParams=locParams)
            if 'PRINT' in sql:
                disp = sql.split("'")[1]
                p('SQL PRINT: ' + disp, "i")
            if len(sql) > 1:
                sql = str(sql) if connObj.isExtractSqlIsOnlySTR else uniocdeStr (sql)
                connObj.exeSQL(sql=sql)
                p (u"FINISH EXEC: %s" % uniocdeStr(sql), "i")

    connObj.connect()
    if str(sqlScript).endswith(".sql") and os.path.isfile(sqlScript):
        with io.open(sqlScript, 'r',  encoding='utf8') as inp:
            __execEachLine(connObj, inp)

    else:
        __execEachLine(connObj, sqlScript)
    connObj.close()

def __split_sql_expressions(text):
    if (isinstance(text, basestring)):
        return [text]
    results = []
    tup = ''
    remove = False
    for line in text:
        if u";" in line:
            lines = line.split (";")
            for l in lines:
                tup+=" "+l
                tup , remove = __split_sql_removeString (tup, remove)
                if len (tup) > 0:
                    results.append (tup)
                tup = ''
        else:
            tup+=" "+line
            tup, remove = __split_sql_removeString(tup, remove)
    if len (tup) > 0:
        results.append(tup)
    return results

def __split_sql_removeString (line, remove=False):

    #if remove:
    #    if "*/" in line:
    #        line = re.sub(r'.*\*\/', '', line)
    #        remove = False
    #    else:
    #        return '', True
    line = line.strip()
    line = re.sub(r'\/\*.*\*\/', '', line)
    line = re.sub(r'--.*', '', line)


    #if "/*" in line:
    #    remove = True
    #    line = re.sub(r'\/\*.*', '', line)
    line = line.replace("\n", " ").replace("\t", " ")
    return line, remove


# -*- coding: utf-8 -*-
# (c) 2017-2019, Tal Shany <tal.shany@biSkilled.com>
#
# This file is part of dingDong
#
# dingDong is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# dingDong is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with dingDong.  If not, see <http://www.gnu.org/licenses/>.

import re
import sqlparse
from collections import OrderedDict
from past.builtins import basestring
from sqlparse.sql       import IdentifierList, Identifier, Parenthesis
from sqlparse.tokens    import Keyword, DML

from dingDong.misc.logger import p
from dingDong.misc.misc import replaceStr, uniocdeStr
from dingDong.config import config

QUERY_COLUMNS_KEY       = '~'
QUERY_NO_TABLE          = '~notFound~'
QUERY_PRE               = '~pre~'
QUERY_POST              = '~post~'

TABLE_ALIAS = 'alias'
TABLE_SCHEMA= 'schema'
TABLE_NAME  = 'tbl'
TABLE_COLUMN= 'column'

## extract_tableAndColumns -->
##      Using extract_tables which return Dictionary : {'Columns': {ColName:ColTargetName....}, 'Tables':{tableName: [Columns ....]}}
##      Using extract_table_identifiers which return list of [(table aliace, table schema, table Name) ...]
def extract_tableAndColumns (sql):
    sql, pre = removeProps (sql=sql)
    tblTupe , columns, sqlPre, sqlPost = extract_tables(sql)


    pre = pre if pre and len(pre)>0 else sqlPre

    ret = {QUERY_PRE:pre, QUERY_POST:sqlPost}

    if QUERY_COLUMNS_KEY in columns:
        ret[QUERY_COLUMNS_KEY] = columns[QUERY_COLUMNS_KEY]
        del columns[QUERY_COLUMNS_KEY]

    # merge table to columns

    tableAliasVsName = {}
    foundTables = []
    for tbl in tblTupe:
        alias       = tbl[0]
        schamenName = tbl[1]
        tableName   = tbl[2]

        aliasTbl = alias if alias is not None and len(alias)>0 else tableName
        tableAliasVsName[aliasTbl] = tableName

        if tableName not in ret:
            ret[tableName] = {TABLE_ALIAS:alias, TABLE_SCHEMA:None if schamenName==tableName else schamenName}

        if columns and len (columns)>0:
            ret[tableName][TABLE_COLUMN] = []

            for col in columns:
                if str(col) in [tableName,alias]:
                    ret[tableName][TABLE_COLUMN].extend  (columns[col])
                    foundTables.append (col)


    for tbl in columns:
        if tbl not in foundTables:
            if QUERY_NO_TABLE not in ret:
                ret[QUERY_NO_TABLE] = {TABLE_ALIAS:None, TABLE_SCHEMA:None,TABLE_COLUMN:[]}
            ret[QUERY_NO_TABLE][TABLE_COLUMN].extend ( columns[tbl] )
    return ret

def removeProps (sql):
    sql = re.sub(r"/\*[^*]*\*+(?:[^*/][^*]*\*+)*/", "", sql)
    sql = re.sub(r"\s+", " ", sql)
    pre = ""

    regSql = r"(.*?select\s+(?:top\s+\d+\s+){0,1}(?:distinct\s+){0,1})(.*)"

    # sqlGrouping = re.search(regexSql, sqlStr, re.IGNORECASE | re.MULTILINE)
    sqlGrouping = re.search(regSql, sql, re.UNICODE | re.MULTILINE | re.S | re.I)
    if sqlGrouping:
        pre = sqlGrouping.group(1)
        sql = '%s %s'  %('SELECT', sqlGrouping.group(2))
    return sql, pre


def extract_tables(sql):
    # let's handle multiple statements in one sql string
    extracted_tables = []
    extracted_last_tables_Tuple = None
    extracted_last_columns_Dic = None
    preSql = ""
    postSql= ""
    # replacements to SQL queries
    sql = replaceStr (sString=sql,findStr="ISNULL (", repStr="ISNULL(", ignoreCase=True,addQuotes=None)
    sql = replaceStr(sString=sql, findStr="CONVERT (",repStr="CONVERT(", ignoreCase=True, addQuotes=None)
    sql = sql.replace("\t"," ")
    statements = list(sqlparse.parse(sql))

    for statement in statements:
        if statement.get_type() != 'UNKNOWN':
            stream = extract_from_part(statement)
            # will get only last table column definistion !
            extracted_last_columns_Dic,preSql,postSql  = extract_select_part(statement)
            extracted_last_tables_Tuple = list (extract_table_identifiers(stream))
            #extracted_tables.append(list(extract_table_identifiers(stream)))
    #return list(itertools.chain(*extracted_tables))
    return (extracted_last_tables_Tuple , extracted_last_columns_Dic,preSql,postSql)

def extract_from_part(parsed):
    from_seen = False
    for item in parsed.tokens:
        if item.is_group:
            for x in extract_from_part(item):
                yield x
        if from_seen:
            if is_subselect(item):
                for x in extract_from_part(item):
                    yield x
            elif item.ttype is Keyword and item.value.upper() in ['ORDER', 'GROUP', 'BY', 'HAVING']:
                from_seen = False
                StopIteration
            else:
                yield item
        if item.ttype is Keyword and item.value.upper() == 'FROM':
            from_seen = True

def is_subselect(parsed):
    if not parsed.is_group:
        return False
    for item in parsed.tokens:
        if item.ttype is DML and item.value.upper() == 'SELECT':
            return True
    return False

def extract_select_part (parsed):
    nonColumnSign       = QUERY_NO_TABLE
    columnList          = []
    columnDic           = {}
    col                 = None
    addToken            = False
    preSql              = ""
    postSql             = ""
    isPost              = False

    for item in parsed.tokens:
        if not addToken and not isPost:
            preSql+=uniocdeStr(item.value)
        if item.ttype is DML and item.value.upper() == 'SELECT':
            addToken = True

        elif item.ttype is Keyword and item.value.upper() == 'FROM':
            addToken = False
            isPost   = True
            postSql += uniocdeStr(item.value)

        elif addToken:
            dicKey  = None
            dicValue= None

            if isinstance(item, IdentifierList):
                for identifier in item.get_identifiers():
                    identifier = str(identifier)
                    srcName = identifier

                    if identifier.lower().rfind(" as ") > 0:
                        srcName = identifier[:identifier.lower().rfind(" as ")].strip()
                        tarName = identifier[identifier.lower().rfind(" as ")+4:].strip()
                    else:
                        tarName = srcName.split(".")
                        tarName = tarName[1] if len(tarName)>1 else tarName[0]

                    columnList.append( (srcName.replace("\n", " ").split(".") , tarName) )
            elif isinstance(item, Identifier):
                item = str(item)
                srcName = item
                if item.lower().find(" as") > 0:
                    srcName = item[:item.lower().find(" as")].strip()
                    tarName = item[item.lower().find(" as")+3:].strip()
                else:
                    tarName = srcName
                columnList.append( (srcName.replace("\n", " ").split(".") , tarName) )
            elif item.value == '*':
                columnList.append( ('*', '*') )

            elif not isPost:
                preSql += uniocdeStr(item.value)
        elif isPost:
            postSql +=uniocdeStr(item.value)

    columnDic[QUERY_COLUMNS_KEY] = columnList
    for tupCol in columnList:
        col     = tupCol[0]
        tarName = tupCol[1]
        if col and len (col) == 1:
            dicKey = nonColumnSign
            dicValue = col[0]
        elif col and len (col) >= 2:
            dicKey = col[0]
            dicValue = col
            if dicKey not in columnDic:
                columnDic[dicKey] = []
        else:
            p("ERROR Loading column identifier, will ignore column %s" % str(col), "i")
            continue

        if dicKey and dicKey not in columnDic:
                columnDic[dicKey] = []
        dicValue = tuple(dicValue) if isinstance(dicValue,list) else dicValue
        dicValue = (dicValue,) if isinstance(dicValue,basestring) else dicValue
        columnDic[dicKey].append ((dicValue,tarName))

    return columnDic, preSql, postSql

def extract_table_identifiers(token_stream):
    for item in token_stream:
        if isinstance(item, IdentifierList):
            for identifier in item.get_identifiers():
                value = ( identifier.get_alias() , identifier._get_first_name(), identifier.get_real_name())
                value = [x.replace('"', '').replace("'", "") if x else None for x in  value]
                value = tuple( value )
                yield value

        elif isinstance(item, Identifier):
            value = (item.get_alias(), item._get_first_name(), item.get_real_name())
            value = [x.replace('"', '').replace("'", "") if x else None for x in value]
            value = tuple(value)
            yield value

""" PARSE QUERY --> GET QUERIES BY USING REGEX """
def querySqriptParserIntoList(sqlScript, getPython=True, removeContent=True, dicProp=None):
    if isinstance(sqlScript, (tuple, list)):
        sqlScript = "".join(sqlScript)
    # return list of sql (splitted by list of params)
    allQueries = __getAllQuery(longStr=sqlScript)
    if getPython:
        allQueries = __getPythonParam(allQueries, mWorld=config.PARSER_SQL_MAIN_KEY)

    if removeContent:
        allQueries = __removeComments(allQueries)

    if dicProp:
        allQueries = __replaceProp(allQueries, dicProp)

    return allQueries

def __getAllQuery(longStr, splitParam=['GO', u';']):
    sqlList = []
    for splP in splitParam:
        if len(sqlList) == 0:
            sqlList = longStr.split(splP)
        else:
            tmpList = list([])
            for sql in sqlList:
                tmpList.extend(sql.split(splP))
            sqlList = tmpList
    return sqlList

def __getPythonParam(queryList, mWorld):
    ret = []
    for query in queryList:
        # Delete all rows which are not relevant
        # Regex : <!popEtl XXXX/>
        # fPythonNot = re.search(r"<!%s([^>].*)/>" % (mWorld), query,flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S)
        # Regex : <!popEtl> ......... </!popEtl>
        reg = re.finditer(r"<!%s(.+?)</!%s>" % (mWorld, mWorld), query,
                          flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S)
        if reg:
            for regRemove in reg:
                query = query.replace(regRemove.group(0), "")

        # Add python queries into return list
        # Regex : <popEtl STRING_NAME> ....... </popEtl>
        # fPython2    = re.search(r"<%s.*/%s>" % (mWorld,mWorld),   query, flags = re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S)

        # Regex : <popEtl STRING_NAME>......</popEtl> --> Take string to the end
        reg = re.finditer(r"<%s(.+?)>(.+?)</%s>" % (mWorld, mWorld), query,flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL | re.S)
        if reg:

            for i, regFind in enumerate(reg):
                pythonSeq = regFind.group(0)
                pythonVar = regFind.group(1).strip()
                querySql = regFind.group(2).strip()

                if i == 0 and regFind.start() > 0:
                    queryStart = query[: query.find(pythonSeq)].strip()
                    if queryStart and len(queryStart) > 0:
                        ret.append((None, queryStart))

                ret.append((pythonVar, querySql))
        else:
            if query and len(query.strip()) > 0:
                ret.append((None, query.strip()))
    return ret

def __removeComments(listQuery, endOfLine='\n'):
    retList = []
    for s in listQuery:
        isTup = False
        if isinstance(s, (tuple, list)):
            pre = s[0].strip() if s[0] else None
            post = s[1].strip()
            isTup = True
        else:
            post = s.strip()

        post = re.sub(r"--.*$", r"", post, flags=re.IGNORECASE | re.MULTILINE | re.UNICODE).replace("--", "")
        post = re.sub(r'\/\*.*\*\/', "", post, flags=re.IGNORECASE | re.MULTILINE | re.UNICODE | re.DOTALL)
        post = re.sub(r"print .*$", r"", post, flags=re.IGNORECASE | re.MULTILINE | re.UNICODE).replace("print ",
                                                                                                        "")

        if endOfLine:
            while len(post) > 1 and post[0:1] == "\n":
                post = post[1:]

            while len(post) > 1 and post[-1:] == "\n":
                post = post[:-1]

        if not post or len(post) == 0:
            continue
        else:
            if isTup:
                retList.append((pre, post,))
            else:
                retList.append(post)

    return retList

def __replaceProp(allQueries, dicProp):
    ret = []
    for query in allQueries:
        if isinstance(query, (list, tuple)):
            pr1 = query[0]
            pr2 = query[1]
        else:
            pr2 = query
        if not pr1 or pr1 and pr1 != "~":
            for prop in dicProp:
                pr2 = (replaceStr(sString=pr2, findStr=prop, repStr=dicProp[prop], ignoreCase=True))

        tupRet = (pr1, pr2,) if isinstance(query, (list, tuple)) else pr2
        ret.append(tupRet)
    return ret

sql = """SELECT Top 1000 t.*, t.CLASS AS INTSugMev, t.CLASS AS KUKU,t.CLASS,
t.CKTXT AS UNSugMevDescr,CASE t.CLASS WHEN '0002' THEN 'BBB' ELSE t.CKTXT END AS Test,
CAST(t.CKTXT AS varchar(100)) As test2,
(SELECT SYSDATE FROM dual)  As gggg
FROM SAPR3.TNP01T t, SAPR3.TNP01 p where t.SPRAS = 'B' and t.CLASS = p.CLASS"""
#xx = extract_tableAndColumns (sql)

#print ("TAL 1234")
#for x in xx:
#    print (x, xx[x])